import contextlib
import json

from dcos import emitting, http, util
from dcos.errors import DCOSException, DCOSHTTPException
from dcoscli import tables

logger = util.get_logger(__name__)
emitter = emitting.FlatEmitter()


def _gib(n):
    return n * pow(2, -30)


def _percentage(dividend, divisor):
    if divisor > 0:
        return dividend / divisor * 100
    return 0


def _fetch_metrics_datapoints(url):
    """Retrieve the metrics data from any `dcos-metrics` endpoint.

    :param url: `dcos-metrics` endpoint
    :type url: str
    :returns: List of metrics datapoints
    :rtype: [dict]
    """
    with contextlib.closing(http.get(url)) as r:

        if r.status_code == 204:
            raise DCOSException('No metrics found')

        if r.status_code != 200:
            raise DCOSHTTPException(r)

        return r.json().get('datapoints', [])


def _get_datapoint(datapoints, name, tags=None):
    """Find a specific datapoint by name and tags

    :param datapoints: a list of datapoints
    :type datapoints: [dict]
    :param name: the name of the required datapoint
    :type name: str
    :param tags: required tags by key and value
    :type tags: dict
    :return: a matching datapoint
    :rtype: dict
    """
    for datapoint in datapoints:
        if datapoint['name'] == name:
            if tags is None:
                return datapoint

            dtags = datapoint.get('tags', {})
            tag_match = True
            for k, v in tags.items():
                tag_match = tag_match and dtags.get(k) == v
            if tag_match:
                return datapoint


def _get_datapoint_value(datapoints, name, tags=None):
    """Safely return only the value from a datapoint, defaulting to 0.

    :param datapoints: a list of datapoints
    :type datapoints: [dict]
    :param name: the name of the required datapoint
    :type name: str
    :param tags: required tags by key and value
    :type tags: dict
    :return: a matching datapoint
    :rtype: float
    """
    datapoint = _get_datapoint(datapoints, name, tags)
    if datapoint is not None:
        return datapoint.get('value', 0.0)

    return 0.0


def _node_summary_json(datapoints):
    """Filters datapoints down to CPU, memory and root disk space fields.

    :param datapoints: a list of datapoints
    :type datapoints: [dict]
    :return: JSON data
    :rtype: str
    """
    summary_datapoints = [
        _get_datapoint(datapoints, 'cpu.total'),
        _get_datapoint(datapoints, 'memory.total'),
        _get_datapoint(datapoints, 'filesystem.capacity.used', {'path': '/'})
    ]
    return json.dumps(summary_datapoints)


def _node_summary_data(datapoints):
    """Extracts CPU, memory and root disk space fields from node datapoints.

    :param datapoints: a list of raw datapoints
    :type datapoints: [dict]
    :return: a dictionary of summary fields
    :rtype: dict
    """

    cpu_used = _get_datapoint_value(datapoints, 'load.1min')
    cpu_used_pc = _get_datapoint_value(datapoints, 'cpu.total')

    mem_total = _get_datapoint_value(datapoints, 'memory.total')
    mem_free = _get_datapoint_value(datapoints, 'memory.free')
    mem_used = mem_total - mem_free
    mem_used_pc = _percentage(mem_used, mem_total)

    disk_total = _get_datapoint_value(
        datapoints, 'filesystem.capacity.total', {'path': '/'})
    disk_free = _get_datapoint_value(
        datapoints, 'filesystem.capacity.used', {'path': '/'})
    disk_used = disk_total - disk_free
    disk_used_pc = _percentage(disk_used, disk_total)

    return {
        'cpu': '{:0.2f} ({:0.2f}%)'.format(cpu_used, cpu_used_pc),
        'mem': '{:0.2f}GiB ({:0.2f}%)'.format(_gib(mem_used), mem_used_pc),
        'disk': '{:0.2f}GiB ({:0.2f}%)'.format(_gib(disk_used), disk_used_pc)
    }


def _task_summary_json(datapoints):
    """Filters datapoints down to CPU, memory and disk space fields.

    :param datapoints: a list of datapoints
    :type datapoints: [dict]
    :return: JSON data
    :rtype: str
    """
    summary_datapoints = [
        _get_datapoint(datapoints, 'cpus.user.time'),
        _get_datapoint(datapoints, 'mem.total'),
        _get_datapoint(datapoints, 'disk.used')
    ]
    return json.dumps(summary_datapoints)


def _task_summary_data(datapoints):
    """Extracts CPU, memory and root disk space fields from task datapoints.

    :param datapoints: a list of raw datapoints
    :type datapoints: [dict]
    :return: a dictionary of summary fields
    :rtype: dict
    """

    cpu_user = _get_datapoint_value(datapoints, 'cpus.user.time')
    cpu_system = _get_datapoint_value(datapoints, 'cpus.system.time')
    cpu_throttled = _get_datapoint_value(datapoints, 'cpus.throttled.time')
    cpu_used = cpu_user + cpu_system
    cpu_total = cpu_used + cpu_throttled
    cpu_used_pc = _percentage(cpu_used, cpu_total)

    mem_total = _get_datapoint_value(datapoints, 'mem.limit')
    mem_used = _get_datapoint_value(datapoints, 'mem.total')
    mem_used_pc = _percentage(mem_used, mem_total)

    disk_total = _get_datapoint_value(datapoints, 'disk.used')
    disk_used = _get_datapoint_value(datapoints, 'disk.limit')
    disk_used_pc = _percentage(disk_used, disk_total)

    return {
        'cpu': '{:0.2f} ({:0.2f}%)'.format(cpu_used, cpu_used_pc),
        'mem': '{:0.2f}GiB ({:0.2f}%)'.format(_gib(mem_used), mem_used_pc),
        'disk': '{:0.2f}GiB ({:0.2f}%)'.format(_gib(disk_used), disk_used_pc)
    }


def _format_datapoints(datapoints):
    """Format raw datapoints for output by making values human-readable
    according to their unit and formatting tags.

    :param datapoints: a list of datapoints
    :type datapoints: [dict]
    :return: a list of formatted datapoints
    :rtype: [dict]
    """

    def _format_tags(tags):
        if tags is None:
            return ''
        pairs = []
        for k, v in tags.items():
            pairs.append('{}: {}'.format(k, v))
        return ', '.join(pairs)

    def _format_value(v, u):
        if u == 'bytes':
            return '{:0.2f}GiB'.format(_gib(v))
        if u == 'percent':
            return '{:0.2f}%'.format(v)
        if type(v) == float:
            return '{:0.2f}'.format(v)
        return v

    formatted_datapoints = []
    for d in datapoints:
        formatted_datapoints.append({
            'name': d['name'],
            'value': _format_value(d['value'], d['unit']),
            'tags': _format_tags(d.get('tags'))
        })

    return formatted_datapoints


def print_node_metrics(url, summary, json_):
    """Retrieve and pretty-print key fields from the `dcos-metrics`' `node`
    endpoint.

    :param url: `dcos-metrics` `node` endpoint
    :type url: str
    :param summary: print summary if true, or all fields if false
    :type summary: bool
    :param json_: print json list if true
    :type json_: bool
    :returns: Process status
    :rtype: int
    """

    datapoints = _fetch_metrics_datapoints(url)

    if summary:
        if json_:
            return emitter.publish(_node_summary_json(datapoints))
        table = tables.metrics_summary_table(_node_summary_data(datapoints))
    else:
        if json_:
            return emitter.publish(datapoints)
        table = tables.metrics_details_table(_format_datapoints(datapoints))

    return emitter.publish(table)


def print_task_metrics(url, app_url, summary, json_):
    """Retrieve and pretty-print fields from the `dcos-metrics`' `containers/id`
    endpoint and `containers/id/app` endpoint.

    :param url: `dcos-metrics` `containers/id` endpoint
    :type url: str
    :param app_url: `dcos-metrics` `containers/id/app` endpoint
    :type app_url: str
    :param json_: print json list if true
    :type json_: bool
    :param summary: print summary if true, or all fields if false
    :type summary: bool
    :return: Process status
    :rtype: int
    """

    datapoints = _fetch_metrics_datapoints(url) + _fetch_metrics_datapoints(
        app_url)

    if summary:
        if json_:
            return emitter.publish(_task_summary_json(datapoints))
        table = tables.metrics_summary_table(_task_summary_data(datapoints))
    else:
        if json_:
            return emitter.publish(datapoints)
        table = tables.metrics_details_table(_format_datapoints(datapoints),
                                             False)

    return emitter.publish(table)



"""
Provides facilities for "Record-IO" encoding of data.
"Record-IO" encoding allows one to encode a sequence
of variable-length records by prefixing each record
with its size in bytes:

5\n
hello
6\n
world!

Note that this currently only supports record lengths
encoded as base 10 integer values with newlines as a
delimiter. This is to provide better language
portability: parsing a base 10 integer is simple. Most
other "Record-IO" implementations use a fixed-size
header of 4 bytes to directly encode an unsigned 32 bit
length.
"""

from dcos.errors import DCOSException


class Encoder(object):
    """Encode an arbitray message type into a 'RecordIO' message.

       This class encapsulates the process of encoding an
       arbitrary message into a 'RecordIO' message. Its
       constructor takes a serialization function of the form
       'serialize(message)'. This serialization function is
       responsible for knowing how to take whatever message type
       is passed to 'encode()' and serializing it to a 'UTF-8'
       encoded byte array.

       Once 'encode(message)' is called, it will use the
       serialization function to convert 'message' into a 'UTF-8'
       encoded byte array, wrap it in a 'RecordIO' frame,
       and return it.

       :param serialize: a function to serialize any message
                         passed to 'encode()' into a 'UTF-8'
                         encoded byte array
       :type serialize: function
    """

    def __init__(self, serialize):
        self.serialize = serialize

    def encode(self, message):
        """Encode a message into 'RecordIO' format.

        :param message: a message to serialize and then wrap in
                        a 'RecordIO' frame.
        :type message: object
        :returns: a serialized message wrapped in a 'RecordIO' frame
        :rtype: bytes
        """

        s = self.serialize(message)

        if not isinstance(s, bytes):
            raise DCOSException("Calling 'serialize(message)' must"
                                " return a 'bytes' object")

        return bytes(str(len(s)) + "\n", "UTF-8") + s


class Decoder(object):
    """Decode a 'RecordIO' message back to an arbitrary message type.

       This class encapsulates the process of decoding a message
       previously encoded with 'RecordIO' back to an arbitrary
       message type. Its constructor takes a deserialization
       function of the form 'deserialize(data)'. This
       deserialization function is responsible for knowing how to
       take a fully constructed 'RecordIO' message containing a
       'UTF-8' encoded byte array and deserialize it back into the
       original message type.

       The 'decode(data)' message takes a 'UTF-8' encoded byte array
       as input and buffers it across subsequent calls to
       construct a set of fully constructed 'RecordIO' messages that
       are decoded and returned in a list.

       :param deserialize: a function to deserialize from 'RecordIO'
                           messages built up by subsequent calls
                           to 'decode(data)'
       :type deserialize: function
    """

    HEADER = 0
    RECORD = 1
    FAILED = 2

    def __init__(self, deserialize):
        self.deserialize = deserialize
        self.state = self.HEADER
        self.buffer = bytes("", "UTF-8")
        self.length = 0

    def decode(self, data):
        """Decode a 'RecordIO' formatted message to its original type.

        :param data: an array of 'UTF-8' encoded bytes that make up a
                      partial 'RecordIO' message. Subsequent calls to this
                      function maintain state to build up a full 'RecordIO'
                      message and decode it
        :type data: bytes
        :returns: a list of deserialized messages
        :rtype: list
        """

        if not isinstance(data, bytes):
            raise DCOSException("Parameter 'data' must of of type 'bytes'")

        if self.state == self.FAILED:
            raise DCOSException("Decoder is in a FAILED state")

        records = []

        for c in data:
            if self.state == self.HEADER:
                if c != ord('\n'):
                    self.buffer += bytes([c])
                    continue

                try:
                    self.length = int(self.buffer.decode("UTF-8"))
                except Exception as exception:
                    self.state = self.FAILED
                    raise DCOSException("Failed to decode length"
                                        "'{buffer}': {error}"
                                        .format(buffer=self.buffer,
                                                error=exception))

                self.buffer = bytes("", "UTF-8")
                self.state = self.RECORD

                # Note that for 0 length records, we immediately decode.
                if self.length <= 0:
                    records.append(self.deserialize(self.buffer))
                    self.state = self.HEADER

            elif self.state == self.RECORD:
                assert self.length
                assert len(self.buffer) < self.length

                self.buffer += bytes([c])

                if len(self.buffer) == self.length:
                    records.append(self.deserialize(self.buffer))
                    self.buffer = bytes("", "UTF-8")
                    self.state = self.HEADER

        return records



import base64
import contextlib
import json
import os
import subprocess
import time
import zipfile

import six
from six.moves import urllib

from dcos import config, http


def exec_command(cmd, env=None, stdin=None):
    """Execute CLI command

    :param cmd: Program and arguments
    :type cmd: [str]
    :param env: Environment variables
    :type env: dict | None
    :param stdin: File to use for stdin
    :type stdin: file
    :returns: A tuple with the returncode, stdout and stderr
    :rtype: (int, bytes, bytes)
    """

    print('CMD: {!r}'.format(cmd))

    process = subprocess.Popen(
        cmd,
        stdin=stdin,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=env)

    # This is needed to get rid of '\r' from Windows's lines endings.
    stdout, stderr = [std_stream.replace(b'\r', b'')
                      for std_stream in process.communicate()]

    # We should always print the stdout and stderr
    print('STDOUT: {}'.format(_truncate(stdout.decode('utf-8'))))
    print('STDERR: {}'.format(_truncate(stderr.decode('utf-8'))))

    return (process.returncode, stdout, stderr)


def _truncate(s, length=8000):
    if len(s) > length:
        return s[:length-3] + '...'
    else:
        return s


def assert_command(
        cmd,
        returncode=0,
        stdout=b'',
        stderr=b'',
        env=None,
        stdin=None):
    """Execute CLI command and assert expected behavior.

    :param cmd: Program and arguments
    :type cmd: list of str
    :param returncode: Expected return code
    :type returncode: int
    :param stdout: Expected stdout
    :type stdout: bytes
    :param stderr: Expected stderr
    :type stderr: bytes
    :param env: Environment variables
    :type env: dict of str to str
    :param stdin: File to use for stdin
    :type stdin: file
    :rtype: None
    """

    returncode_, stdout_, stderr_ = exec_command(cmd, env, stdin)

    assert returncode_ == returncode, (returncode_, returncode)
    assert stdout_ == stdout, (stdout_, stdout)
    assert stderr_ == stderr, (stderr_, stderr)


def delete_zk_nodes():
    """Delete Zookeeper nodes that were created during the tests

    :rtype: None
    """

    for znode in ['universe', 'dcos-service-cassandra', 'chronos']:
        delete_zk_node(znode)


def delete_zk_node(znode):
    """Delete Zookeeper node

    :param znode: znode to delete
    :type znode: str
    :rtype: None
    """

    dcos_url = config.get_config_val('core.dcos_url')
    znode_url = urllib.parse.urljoin(
        dcos_url,
        '/exhibitor/exhibitor/v1/explorer/znode/{}'.format(znode))
    http.delete(znode_url)


def assert_lines(cmd, num_lines, greater_than=False):
    """ Assert stdout contains the expected number of lines

    :param cmd: program and arguments
    :type cmd: [str]
    :param num_lines: expected number of lines for stdout
    :type num_lines: int
    :param greater_than: if True assume there may be at least num_lines or more
    :type greater_than: bool
    :rtype: None
    """

    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 0
    assert stderr == b''
    lines = len(stdout.decode('utf-8').split('\n')) - 1
    if greater_than:
        assert lines >= num_lines
        return
    assert lines == num_lines


def file_json_ast(path):
    """Returns the JSON AST parsed from file
    :param path: path to file
    :type path: str
    :returns: parsed JSON AST
    """
    with open(path) as f:
        return json.load(f)


def json_ast_format(ast):
    """Returns the given JSON AST formatted as bytes

    :param ast: JSON AST
    :returns: formatted JSON
    :rtype: bytes
    """
    return six.b(
        json.dumps(ast,
                   sort_keys=True,
                   indent=2,
                   separators=(',', ': '))) + b'\n'


def fetch_valid_json(cmd):
    """Assert stdout contains valid JSON

    :param cmd: program and arguments
    :type cmd: [str]
    :returns: parsed JSON AST
    """
    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 0
    assert stderr == b''
    try:
        return json.loads(stdout.decode('utf-8'))
    except json.JSONDecodeError:
        error_text = 'Command "{}" returned invalid JSON'.format(' '.join(cmd))
        raise Exception(error_text)


def file_json(path):
    """ Returns formatted json from file

    :param path: path to file
    :type path: str
    :returns: formatted json
    :rtype: bytes
    """
    return json_ast_format(file_json_ast(path))


@contextlib.contextmanager
def update_config(name, value, env=None):
    """ Context manager for altering config for tests

    :param key: <key>
    :type key: str
    :param value: <value>
    :type value: str
    ;param env: env vars
    :type env: dict
    :rtype: None
    """

    returncode, stdout, _ = exec_command(
        ['dcos', 'config', 'show', name], env)

    # when we change the dcos_url we remove the acs_token
    # we need to also restore the token if this occurs
    token = None
    if name == "core.dcos_url":
        returncode_, token_val, _ = exec_command(
            ['dcos', 'config', 'show', "core.dcos_acs_token"], env)
        if returncode_ == 0:
            token = token_val.decode('utf-8').strip()

    result = None
    # config param already exists
    if returncode == 0:
        result = json.loads('"' + stdout.decode('utf-8').strip() + '"')

    # if we are setting a value
    if value is not None:
        config_set(name, value, env)
    # only unset if the config param already exists
    elif result is not None:
        config_unset(name, env)

    try:
        yield
    finally:
        # return config to previous state
        if result is not None:
            config_set(name, result, env)
        else:
            exec_command(['dcos', 'config', 'unset', name], env)

    if token:
        config_set("core.dcos_acs_token", token, env)


def popen_tty(cmd):
    """Open a process with stdin connected to a pseudo-tty.  Returns a

    :param cmd: command to run
    :type cmd: str
    :returns: (Popen, master) tuple, where master is the master side
       of the of the tty-pair.  It is the responsibility of the caller
       to close the master fd, and to perform any cleanup (including
       waiting for completion) of the Popen object.
    :rtype: (Popen, int)

    """

    import pty
    master, slave = pty.openpty()
    proc = subprocess.Popen(cmd,
                            stdin=slave,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            preexec_fn=os.setsid,
                            close_fds=True,
                            shell=True)
    os.close(slave)

    return (proc, master)


def ssh_output(cmd):
    """ Runs an SSH command and returns the stdout/stderr/returncode.

    :param cmd: command to run
    :type cmd: str
    :rtype: (str, str, int)
    """

    print('SSH COMMAND: {}'.format(cmd))

    # ssh must run with stdin attached to a tty
    proc, master = popen_tty(cmd)

    # wait for the ssh connection
    time.sleep(5)

    proc.poll()
    returncode = proc.returncode

    # kill the whole process group
    try:
        os.killpg(os.getpgid(proc.pid), 15)
    except OSError:
        pass

    os.close(master)
    stdout, stderr = proc.communicate()

    print('SSH STDOUT: {}'.format(stdout.decode('utf-8')))
    print('SSH STDERR: {}'.format(stderr.decode('utf-8')))

    return stdout, stderr, returncode


def config_set(key, value, env=None):
    """ dcos config set <key> <value>

    :param key: <key>
    :type key: str
    :param value: <value>
    :type value: str
    ;param env: env vars
    :type env: dict
    :rtype: None
    """
    returncode, stdout, _ = exec_command(
        ['dcos', 'config', 'set', key, value],
        env=env)

    assert returncode == 0
    assert stdout == b''


def config_unset(key, env=None):
    """ dcos config unset <key>

    :param key: <key>
    :type key: str
    :param env: env vars
    :type env: dict
    :rtype: None
    """

    cmd = ['dcos', 'config', 'unset', key]

    returncode, stdout, stderr = exec_command(cmd, env=env)

    assert returncode == 0
    assert stdout == b''


def base64_to_dict(byte_string):
    """
    :param byte_string: base64 encoded string
    :type byte_string: str
    :return: python dictionary decoding of byte_string
    :rtype dict
    """
    return json.loads(base64.b64decode(byte_string).decode('utf-8'))


def zip_contents_as_json(path, inner_file):
    with zipfile.ZipFile(path) as zip_file:
        inner_file_contents = zip_file.read(inner_file).decode()
    return json.loads(inner_file_contents)



import os

import docopt

import dcoscli
from dcos import auth, cmds, config, emitting, http, util
from dcos.errors import DCOSException, DefaultError
from dcoscli import tables
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import decorate_docopt_usage


emitter = emitting.FlatEmitter()
logger = util.get_logger(__name__)


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("auth"),
        argv=argv,
        version='dcos-auth version {}'.format(dcoscli.version))

    http.silence_requests_warnings()

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: all the supported commands
    :rtype: list of dcos.cmds.Command
    """

    return [
        cmds.Command(
            hierarchy=['auth', 'list-providers'],
            arg_keys=['--json'],
            function=_list_providers),

        cmds.Command(
            hierarchy=['auth', 'login'],
            arg_keys=['--password', '--password-env', '--password-file',
                      '--provider', '--username', '--private-key'],
            function=_login),

        cmds.Command(
            hierarchy=['auth', 'logout'],
            arg_keys=[],
            function=_logout),

        cmds.Command(
            hierarchy=['auth'],
            arg_keys=['--info'],
            function=_info),
    ]


def _info(info):
    """
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: process status
    :rtype: int
    """

    emitter.publish(default_command_info("auth"))
    return 0


def _list_providers(json_):
    """
    :returns: providers available for configured cluster
    :rtype: dict
    """

    providers = auth.get_providers()
    if providers or json_:
        emitting.publish_table(
            emitter, providers, tables.auth_provider_table, json_)
    else:
        raise DCOSException("No providers configured for your cluster")


def _get_password(password_str, password_env, password_file):
    """
    Get password for authentication

    :param password_str: password
    :type password_str: str
    :param password_env: name of environment variable with password
    :type password_env: str
    :param password_file: path to file with password
    :type password_file: bool
    :returns: password or None if no password specified
    :rtype: str | None
    """

    password = None
    if password_str:
        password = password_str
    elif password_env:
        password = os.environ.get(password_env)
        if password is None:
            msg = "Environment variable specified [{}] does not exist"
            raise DCOSException(msg.format(password_env))
    elif password_file:
        password = util.read_file_secure(password_file)
    return password


def _login(password_str, password_env, password_file,
           provider, username, key_path):
    """
    :param password_str: password
    :type password_str: str
    :param password_env: name of environment variable with password
    :type password_env: str
    :param password_file: path to file with password
    :type password_file: bool
    :param provider: name of provider to authentication with
    :type provider: str
    :param username: username
    :type username: str
    :param key_path: path to file with private key
    :type param: str
    :rtype: int
    """

    dcos_url = config.get_config_val("core.dcos_url")
    if dcos_url is None:
        msg = ("Please provide the url to your DC/OS cluster: "
               "`dcos config set core.dcos_url`")
        raise DCOSException(msg)

    # every call to login will generate a new token if applicable
    _logout()

    login(dcos_url, password_str, password_env, password_file,
          provider, username, key_path)

    emitter.publish("Login successful!")
    return 0


def _trigger_client_method(
        provider, provider_info, username=None, password=None, key_path=None):
    """
    Trigger client method for authentication type user requested

    :param provider: provider_id requested by user
    :type provider: str
    :param provider_info: info about auth type defined by provider
    :param provider_info: dict
    :param username: username
    :type username: str
    :param password: password
    :type password: str
    :param key_path: path to file with service key
    :type param: str
    :rtype: None
    """

    client_method = provider_info.get("client-method")
    dcos_url = config.get_config_val("core.dcos_url")

    if client_method == "browser-prompt-authtoken":
        auth.browser_prompt_auth(dcos_url, provider_info)
    elif client_method == "browser-prompt-oidcidtoken-get-authtoken":
        auth.oidc_implicit_flow_auth(dcos_url)
    elif client_method == "dcos-usercredential-post-receive-authtoken" or \
            client_method == "dcos-credential-post-receive-authtoken":
        if not username or not password:
            msg = "Please specify username and password for provider [{}]"
            raise DCOSException(msg.format(provider))
        auth.dcos_uid_password_auth(dcos_url, username, password)
    elif client_method == "dcos-servicecredential-post-receive-authtoken":
        if not username or not key_path:
            msg = "Please specify username and service key for provider [{}]"
            raise DCOSException(msg.format(provider))
        auth.servicecred_auth(dcos_url, username, key_path)
    else:
        msg = "Authentication by provider [{}] is not supported by this CLI"
        raise DCOSException(msg.format(provider))


def _logout():
    """
    Logout the user from dcos acs auth or oauth

    :returns: process status
    :rtype: int
    """

    if config.get_config_val("core.dcos_acs_token") is not None:
        config.unset("core.dcos_acs_token")
    return 0


def login(dcos_url, password_str, password_env, password_file,
          provider, username, key_path):
    """
    :param dcos_url: URL of DC/OS cluster
    :type dcos_url: str
    :param password_str: password
    :type password_str: str
    :param password_env: name of environment variable with password
    :type password_env: str
    :param password_file: path to file with password
    :type password_file: bool
    :param provider: name of provider to authentication with
    :type provider: str
    :param username: username
    :type username: str
    :param key_path: path to file with private key
    :type param: str
    :rtype: int
    """

    password = _get_password(password_str, password_env, password_file)
    if provider is None:
        if username and password:
            auth.dcos_uid_password_auth(dcos_url, username, password)
        elif username and key_path:
            auth.servicecred_auth(dcos_url, username, key_path)
        else:
            try:
                providers = auth.get_providers()
                # Let users know if they have non-default providers configured
                # This is a weak check, we should check default versions per
                # DC/OS version since defaults will change. jj
                if len(providers) > 2:
                    msg = ("\nYour cluster has several authentication "
                           "providers enabled. Run `dcos auth "
                           "list-providers` to see all providers and `dcos "
                           "auth login --provider <provider-id>` to "
                           "authenticate with a specific provider\n")
                    emitter.publish(DefaultError(msg))
            except DCOSException:
                pass
            finally:
                auth.header_challenge_auth(dcos_url)
    else:
        providers = auth.get_providers()
        if providers.get(provider):
            _trigger_client_method(
                provider, providers[provider], username, password, key_path)
        else:
            msg = "Provider [{}] not configured on your cluster"
            raise DCOSException(msg.format(provider))

    return 0



import os
import subprocess


class Subproc():
    """Represents a wrapper for subprocess
    """

    def __init__(self):
        self._env = os.environ.copy()
        lib_path = 'LD_LIBRARY_PATH'
        lib_path_value = self._env.get(lib_path + '_ORIG')
        # remove pyinstaller overriding `LD_LIBRARY_PATH`
        # remove with stable release of fix:
        # https://github.com/pyinstaller/pyinstaller/pull/2148
        if lib_path_value is not None:
            self._env[lib_path] = lib_path_value
        elif self._env.get(lib_path):
            del self._env[lib_path]

        # Setuptools overrides path to executable from virtualenv,
        # modify this so we can specify a different path
        launcher = '__PYVENV_LAUNCHER__'
        pyvenv_launcher = self._env.get(launcher)
        if pyvenv_launcher is not None:
            del self._env[launcher]

    def check_output(self, args, stdin=None, stderr=None, shell=False):
        """
        call subprocess.check_ouput with modified environment
        """

        return subprocess.check_output(
            args,
            stdin=stdin,
            stderr=stderr,
            shell=shell,
            env=self._env)

    def popen(self, args, stdin=None, stdout=None, stderr=None, shell=False):
        """
        call subprocess.Popen with modified environment
        """

        return subprocess.Popen(
            args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            shell=shell,
            env=self._env)

    def call(self, args, stdin=None, stdout=None, stderr=None, shell=False):
        """
        call subprocess.call with modified environment
        """

        return subprocess.call(
            args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            shell=shell,
            env=self._env)



import pytest
import requests
from mock import create_autospec, patch

from dcos import auth
from dcos.errors import DCOSException


def test_get_auth_scheme():
    _get_auth_scheme({'www-authenticate': 'acsjwt'}, scheme='acsjwt')
    _get_auth_scheme({'www-authenticate': 'oauthjwt'}, scheme='oauthjwt')
    _get_auth_scheme({}, scheme=None)

    msg = ("Server responded with an HTTP 'www-authenticate' field of "
           "'foobar', DC/OS only supports ['oauthjwt', 'acsjwt']")
    _get_auth_scheme_exception({'www-authenticate': 'foobar'}, msg)


def _get_auth_scheme(header, scheme):
    with patch('requests.Response') as mock:
        mock.headers = header
        auth_scheme = auth._get_auth_scheme(mock)
        assert auth_scheme == scheme


def _get_auth_scheme_exception(header, err_msg):
    with patch('requests.Response') as mock:
        mock.headers = header
        with pytest.raises(DCOSException) as e:
            auth._get_auth_scheme(mock)

    assert str(e.value) == err_msg


@patch('dcos.http._request')
@patch('dcos.config.set_val')
def test_get_dcostoken_by_post_with_creds(config, req):
    creds = {"foobar"}
    resp = create_autospec(requests.Response)
    resp.status_code = 200
    resp.json.return_value = {"token": "foo"}
    req.return_value = resp

    auth._get_dcostoken_by_post_with_creds("http://url", creds)
    req.assert_called_with(
        "post", "http://url/acs/api/v1/auth/login", json=creds)
    config.assert_called_with("core.dcos_acs_token", "foo")


@patch('dcos.http._request')
@patch('dcos.auth._get_dcostoken_by_oidc_implicit_flow')
@patch('dcos.auth._get_dcostoken_by_dcos_uid_password_auth')
def test_header_challenge_auth(cred_auth, oidc_auth, req):
    resp = create_autospec(requests.Response)
    resp.status_code = 401
    resp.headers = {"www-authenticate": "oauthjwt"}
    req.return_value = resp

    auth.header_challenge_auth("url")
    oidc_auth.assert_called_once()

    resp2 = create_autospec(requests.Response)
    resp2.status_code = 401
    resp2.headers = {"www-authenticate": "acsjwt"}
    req.return_value = resp2

    auth.header_challenge_auth("url")
    cred_auth.assert_called_once()


@patch('dcos.http.get')
@patch('dcos.config.get_config_val')
def test_get_providers(config, req_mock):
    resp = create_autospec(requests.Response)
    resp.return_value = {}
    req_mock.return_value = resp
    config.return_value = "http://localhost"

    auth.get_providers()
    req_mock.assert_called_with(
        "http://localhost/acs/api/v1/auth/providers")

    # test url construction valid with trailing slash
    config.return_value = "http://localhost/"

    auth.get_providers()
    req_mock.assert_called_with(
        "http://localhost/acs/api/v1/auth/providers")


@patch('dcos.http._request')
@patch('dcos.config.get_config_val')
def test_get_providers_errors(config, req):
    config.return_value = "http://localhost"

    resp = create_autospec(requests.Response)
    resp.status_code = 404
    req.return_value = resp

    with pytest.raises(DCOSException) as e:
        auth.get_providers()

    err_msg = "This command is not supported for your cluster"
    assert str(e.value) == err_msg



import datetime

import mock
import pytz

from dcos.mesos import Slave
from dcoscli import tables

from ..fixtures.auth_provider import auth_provider_fixture
from ..fixtures.clusters import cluster_list_fixture
from ..fixtures.marathon import (app_fixture, app_task_fixture,
                                 deployment_fixture_app_post_pods,
                                 deployment_fixture_app_pre_pods,
                                 deployment_fixture_pod,
                                 group_fixture, pod_list_fixture,
                                 pod_list_without_instances_fixture,
                                 pod_list_without_spec_version_fixture)
from ..fixtures.metrics import (agent_metrics_node_details_fixture,
                                agent_metrics_node_summary_fixture,
                                agent_metrics_task_details_fixture)
from ..fixtures.metronome import (job_history_fixture, job_list_fixture,
                                  job_run_fixture, job_schedule_fixture)
from ..fixtures.node import slave_fixture
from ..fixtures.package import package_fixture, search_result_fixture
from ..fixtures.service import framework_fixture
from ..fixtures.task import browse_fixture, task_fixture


def test_task_table():
    task = task_fixture()
    task.user = mock.Mock(return_value='root')
    slave = Slave({"hostname": "mock-hostname"}, None, None)
    task.slave = mock.Mock(return_value=slave)
    _test_table(tables.task_table,
                [task],
                'tests/unit/data/task.txt')


def test_app_table():
    apps = [app_fixture()]
    deployments = []
    table = tables.app_table(apps, deployments)
    with open('tests/unit/data/app.txt') as f:
        assert str(table) == f.read()


def test_deployment_table_app_pre_pods():
    _test_table(tables.deployment_table,
                [deployment_fixture_app_pre_pods()],
                'tests/unit/data/deployment/app.txt')


def test_deployment_table_app_post_pods():
    _test_table(tables.deployment_table,
                [deployment_fixture_app_post_pods()],
                'tests/unit/data/deployment/app.txt')


def test_deployment_table_pod():
    _test_table(tables.deployment_table,
                [deployment_fixture_pod()],
                'tests/unit/data/deployment/pod.txt')


def test_app_task_table():
    _test_table(tables.app_task_table,
                [app_task_fixture()],
                'tests/unit/data/app_task.txt')


def test_service_table():
    _test_table(tables.service_table,
                [framework_fixture()],
                'tests/unit/data/service.txt')


def test_group_table():
    _test_table(tables.group_table,
                [group_fixture()],
                'tests/unit/data/group.txt')


def test_auth_providers_table():
    _test_table(tables.auth_provider_table,
                auth_provider_fixture(),
                'tests/unit/data/auth_provider.txt')


def test_job_list_table():
    _test_table(tables.job_table,
                job_list_fixture(),
                'tests/unit/data/job_list.txt')


def test_job_runs_table():
    _test_table(tables.job_runs_table,
                job_run_fixture(),
                'tests/unit/data/job_runs.txt')


def test_job_history_table():
    _test_table(tables.job_history_table,
                job_history_fixture(),
                'tests/unit/data/job_history.txt')


def test_job_schedule_table():
    _test_table(tables.schedule_table,
                job_schedule_fixture(),
                'tests/unit/data/job_schedule.txt')


def test_pod_table():
    _test_table(tables.pod_table,
                pod_list_fixture(),
                'tests/unit/data/pod.txt')


def test_pod_table_without_instances():
    _test_table(tables.pod_table,
                [pod_list_without_instances_fixture()],
                'tests/unit/data/pod_without_instances.txt')


def test_pod_table_without_spec_version():
    _test_table(tables.pod_table,
                [pod_list_without_spec_version_fixture()],
                'tests/unit/data/pod_without_spec_version.txt')


def test_package_table():
    _test_table(tables.package_table,
                [package_fixture()],
                'tests/unit/data/package.txt')


def test_package_search_table():
    _test_table(tables.package_search_table,
                search_result_fixture(),
                'tests/unit/data/package_search.txt')


def test_node_table():
    _test_table(tables.node_table,
                [slave_fixture()],
                'tests/unit/data/node.txt')


def test_clusters_tables():
    _test_table(tables.clusters_table,
                [cluster_list_fixture()],
                'tests/unit/data/cluster.txt')


def test_ls_long_table():
    with mock.patch('dcoscli.tables._format_unix_timestamp',
                    lambda ts: datetime.datetime.fromtimestamp(
                        ts, pytz.utc).strftime('%b %d %H:%M')):
        _test_table(tables.ls_long_table,
                    browse_fixture(),
                    'tests/unit/data/ls_long.txt')


def test_metrics_summary_table():
    _test_table(tables.metrics_summary_table,
                agent_metrics_node_summary_fixture(),
                'tests/unit/data/metrics_summary.txt')


def test_metrics_details_table():
    _test_table(tables.metrics_details_table,
                agent_metrics_node_details_fixture(),
                'tests/unit/data/metrics_node_details.txt')


def test_metrics_details_no_tags_table():
    # Convenience wrapper to pass a second param
    def _task_details_fn(datapoints):
        return tables.metrics_details_table(datapoints, False)
    _test_table(_task_details_fn, agent_metrics_task_details_fixture(),
                'tests/unit/data/metrics_task_details.txt')


def _test_table(table_fn, fixture_fn, path):
    table = table_fn(fixture_fn)
    with open(path) as f:
        assert str(table) == f.read().strip('\n')



import json

from .helpers.common import (assert_command, delete_zk_nodes, exec_command,
                             file_json, file_json_ast)
from .helpers.marathon import (app, group, pod, show_app, show_group,
                               show_pod, start_app, watch_all_deployments)
from .helpers.package import (package, setup_universe_server,
                              teardown_universe_server)
from .helpers.service import get_services, wait_for_service


def setup_module(module):
    setup_universe_server()


def teardown_module(module):
    teardown_universe_server()
    delete_zk_nodes()


_ZERO_INSTANCE_APP = 'tests/data/marathon/apps/zero_instance_sleep.json'


def test_add_app():
    app_id = 'zero-instance-app'
    with app(_ZERO_INSTANCE_APP, app_id):
        show_app('zero-instance-app')


def test_restarting_app():
    app_id = 'zero-instance-app'
    with app(_ZERO_INSTANCE_APP, app_id):
        start_app(app_id, 3)
        watch_all_deployments()
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'app', 'restart', app_id])

        assert returncode == 0
        assert stdout.decode().startswith('Created deployment ')
        assert stderr == b''


def test_add_group():
    group_id = 'test-group'
    with group('tests/data/marathon/groups/good.json', group_id):
        show_group(group_id)


def test_update_group():
    group_app = 'tests/data/marathon/groups/good.json'
    with group(group_app, 'test-group'):
        newapp = json.dumps([{"id": "appadded", "cmd": "sleep 0"}])
        appjson = "apps={}".format(newapp)
        returncode, stdout, stderr = exec_command(
            ['dcos', 'marathon', 'group', 'update', 'test-group/sleep',
                appjson])

        assert returncode == 0
        assert stdout.decode().startswith('Created deployment ')
        assert stderr == b''

        watch_all_deployments()
        show_app('test-group/sleep/appadded')


def test_add_pod():
    pod_id = 'good-pod'
    with pod('tests/data/marathon/pods/good.json', pod_id):
        expected = file_json_ast('tests/data/marathon/pods/good_status.json')
        show_pod(pod_id, expected)


def test_repo_list():
    repo_list = file_json(
        'tests/data/package/json/test_repo_list.json')
    assert_command(
        ['dcos', 'package', 'repo', 'list', '--json'], stdout=repo_list)


def test_package_describe():
    stdout = file_json(
        'tests/data/package/json/test_describe_marathon.json')

    returncode_, stdout_, stderr_ = exec_command(
        ['dcos', 'package', 'describe', 'marathon'])

    assert returncode_ == 0
    output = json.loads(stdout_.decode('utf-8'))
    assert output == json.loads(stdout.decode('utf-8'))
    assert stderr_ == b''


def test_install():
    with package('chronos', deploy=True, args=[]):
        watch_all_deployments()
        wait_for_service('chronos')

    services = get_services(args=['--inactive'])
    assert len([service for service in services
                if service['name'] == 'chronos']) == 0



DCOS_DIR = ".dcos"
"""DC/OS data directory. Can store subcommands and the config file."""

DCOS_DIR_ENV = 'DCOS_DIR'
"""Name of the environment variable pointing to the DC/OS data directory"""

DCOS_CLUSTERS_SUBDIR = "clusters"
"""Name of the subdirectory containing the configuration of all configured
clusters"""

DCOS_CLUSTER = 'DCOS_CLUSTER'
"""Name of the environment variable pointing to the DC/OS cluster id."""

DCOS_CLUSTER_ATTACHED_FILE = "attached"
"""Name of the file indicating the current attached cluster"""

DCOS_SUBCOMMAND_ENV_SUBDIR = 'env'
"""In a package's directory, this is the cli contents subdirectory."""

DCOS_SUBCOMMAND_SUBDIR = 'subcommands'
"""Name of the subdirectory that contains all of the subcommands. This is
relative to the location of the executable."""

DCOS_CONFIG_ENV = 'DCOS_CONFIG'
"""Name of the environment variable pointing to the DC/OS config."""

DCOS_LOG_LEVEL_ENV = 'DCOS_LOG_LEVEL'
"""Name of the environment variable for the DC/OS log level"""

DCOS_DEBUG_ENV = 'DCOS_DEBUG'
"""Name of the environment variable to enable DC/OS debug messages"""

DCOS_PAGER_COMMAND_ENV = 'PAGER'
"""Command to use to page long command output (e.g. 'less -R')"""

PATH_ENV = 'PATH'
"""Name of the environment variable pointing to the executable directories."""

DCOS_COMMAND_PREFIX = 'dcos-'
"""Prefix for all the DC/OS CLI commands."""

VALID_LOG_LEVEL_VALUES = ['debug', 'info', 'warning', 'error', 'critical']
"""List of all the supported log level values for the CLIs"""



import collections
import copy
import json
import os

import pkg_resources
import toml

from dcos import constants, jsonitem, util
from dcos.errors import DCOSException

logger = util.get_logger(__name__)


def uses_deprecated_config():
    """Returns True if the configuration for the user's CLI
    is the deprecated 'global' config instead of the cluster
    specific config
    """

    global_config = get_global_config_path()
    cluster_config = get_clusters_path()
    return not os.path.exists(cluster_config) and os.path.exists(global_config)


def get_global_config_path():
    """Returns the path to the deprecated global DCOS config file.

    :returns: path to the DCOS config file
    :rtype: str
    """

    default_path = os.path.join(get_config_dir_path(), "dcos.toml")
    return os.environ.get(constants.DCOS_CONFIG_ENV, default_path)


def get_global_config(mutable=False):
    """Returns the deprecated global DCOS config file

    :param mutable: True if the returned Toml object should be mutable
    :type mutable: boolean
    :returns: Configuration object
    :rtype: Toml | MutableToml
    """

    return load_from_path(get_global_config_path(), mutable)


def get_attached_cluster_path():
    """
    The attached cluster is denoted by a file named "attached" in one of the
    cluster directories. Ex: $DCOS_DIR/clusters/CLUSTER_ID/attached

    :returns: path to the director of the attached cluster
    :rtype: str | None
    """

    path = get_clusters_path()
    if not os.path.exists(path):
        return None

    attached = None
    clusters = os.listdir(get_clusters_path())
    cluster_envvar = os.environ.get(constants.DCOS_CLUSTER)

    for c in clusters:
        cluster_path = os.path.join(path, c)
        name = get_config_val("cluster.name",
                              load_from_path(
                                  os.path.join(cluster_path, "dcos.toml")))
        if cluster_envvar is not None \
           and (cluster_envvar == c or cluster_envvar == name):
            return cluster_path
        if os.path.exists(os.path.join(
                cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE)):
            attached = cluster_path

    if attached is not None:
        return attached

    # if only one cluster, set as attached
    if len(clusters) == 1:
        cluster_path = os.path.join(path, clusters[0])
        util.ensure_file_exists(os.path.join(
            cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE))
        return cluster_path

    return None


def get_clusters_path():
    """
    :returns: path to the directory of cluster configs
    :rtype: str
    """

    return os.path.join(get_config_dir_path(), constants.DCOS_CLUSTERS_SUBDIR)


def get_config_path():
    """Returns the path to the DCOS config file of the attached cluster.
    If still using "global" config return that toml instead

    :returns: path to the DCOS config file
    :rtype: str
    """

    if uses_deprecated_config():
        return get_global_config_path()
    else:
        cluster_path = get_attached_cluster_path()
        return os.path.join(cluster_path, "dcos.toml")


def get_config_dir_path():
    """
    :returns: path to the DC/OS data directory
    :rtype: str
    """

    config_dir = os.environ.get(constants.DCOS_DIR_ENV) or \
        os.path.join("~", constants.DCOS_DIR)
    return os.path.expanduser(config_dir)


def get_config(mutable=False):
    """Returns the DCOS configuration object and creates config file is None
    found and `DCOS_CONFIG` set to default value. Only use to get the config,
    not to resolve a specific config parameter. This should be done with
    `get_config_val`.

    :param mutable: True if the returned Toml object should be mutable
    :type mutable: boolean
    :returns: Configuration object
    :rtype: Toml | MutableToml
    """

    cluster_path = get_attached_cluster_path()
    if cluster_path is None:
        if uses_deprecated_config():
            return get_global_config(mutable)

        msg = ("No cluster is attached. "
               "Please run `dcos cluster attach <cluster-name>`")
        raise DCOSException(msg)

    util.ensure_dir_exists(os.path.dirname(cluster_path))

    path = os.path.join(cluster_path, "dcos.toml")
    return load_from_path(path, mutable)


def get_config_val_envvar(name, config=None):
    """Returns a tuple of the config value for the specified key and
    the name of any environment variable which overwrote the value
    from the configuration.  If no environment variable applies, None
    is returned for the environment variable name.
    Looks for corresponding environment variable first, and if it
    doesn't exist, uses the config value.
    - "core" properties get resolved to env variable DCOS_SUBKEY. With the
    exception of subkeys that already start with DCOS, in which case we look
    for SUBKEY first, and "DCOS_SUBKEY" second, and finally the config value.
    - everything else gets resolved to DCOS_SECTION_SUBKEY

    :param name: name of paramater
    :type name: str
    :param config: config
    :type config: Toml
    :returns: value of 'name' parameter
    :rtype: (str | None, str | None)
    """

    if config is None:
        config = get_config()

    section, subkey = split_key(name.upper())

    env_var = None
    if section == "CORE":
        if subkey.startswith("DCOS") and os.environ.get(subkey):
            env_var = subkey
        else:
            env_var = "DCOS_{}".format(subkey)
    else:
        env_var = "DCOS_{}_{}".format(section, subkey)

    return os.environ.get(env_var) or config.get(name), env_var or None


def get_config_val(name, config=None):
    """Returns the config value for the specified key. Looks for corresponding
    environment variable first, and if it doesn't exist, uses the config value.
    - "core" properties get resolved to env variable DCOS_SUBKEY. With the
    exception of subkeys that already start with DCOS, in which case we look
    for SUBKEY first, and "DCOS_SUBKEY" second, and finally the config value.
    - everything else gets resolved to DCOS_SECTION_SUBKEY

    :param name: name of paramater
    :type name: str
    :param config: config
    :type config: Toml
    :returns: value of 'name' parameter
    :rtype: str | None
    """

    val, _ = get_config_val_envvar(name, config)
    return val


def missing_config_exception(keys):
    """ DCOSException for a missing config value

    :param keys: keys in the config dict
    :type keys: [str]
    :returns: DCOSException
    :rtype: DCOSException
    """

    msg = '\n'.join(
        'Missing required config parameter: "{0}".'.format(key) +
        '  Please run `dcos config set {0} <value>`.'.format(key)
        for key in keys)
    return DCOSException(msg)


def set_val(name, value, config_path=None):
    """
    :param name: name of paramater
    :type name: str
    :param value: value to set to paramater `name`
    :type param: str
    :param config_path: path to config to use
    :type config_path: str
    :returns: Toml config, message of change
    :rtype: Toml, str
    """

    if config_path:
        toml_config = load_from_path(config_path, True)
    else:
        toml_config = get_config(True)

    section, subkey = split_key(name)

    config_schema = get_config_schema(section)

    new_value = jsonitem.parse_json_value(subkey, value, config_schema)

    toml_config_pre = copy.deepcopy(toml_config)
    if section not in toml_config_pre._dictionary:
        toml_config_pre._dictionary[section] = {}

    value_exists = name in toml_config
    old_value = toml_config.get(name)

    # remove token when core.dcos_url is changed
    token_unset = False
    if value_exists and old_value != new_value and name == "core.dcos_url":
        token_unset = bool(toml_config.pop("core.dcos_acs_token", False))

    toml_config[name] = new_value

    check_config(toml_config_pre, toml_config, section)

    save(toml_config, config_path)

    msg = "[{}]: ".format(name)
    if name == "core.dcos_acs_token":
        if not value_exists:
            msg += "set"
        elif old_value == new_value:
            msg += "already set to that value"
        else:
            msg += "changed"
    elif not value_exists:
        msg += "set to '{}'".format(new_value)
    elif old_value == new_value:
        msg += "already set to '{}'".format(old_value)
    else:
        msg += "changed from '{}' to '{}'".format(old_value, new_value)

    if token_unset:
        msg += "\n[core.dcos_acs_token]: removed"

    return toml_config, msg


def load_from_path(path, mutable=False):
    """Loads a TOML file from the path

    :param path: Path to the TOML file
    :type path: str
    :param mutable: True if the returned Toml object should be mutable
    :type mutable: boolean
    :returns: Map for the configuration file
    :rtype: Toml | MutableToml
    """

    util.ensure_file_exists(path)
    util.enforce_file_permissions(path)
    with util.open_file(path, 'r') as config_file:
        try:
            toml_obj = toml.loads(config_file.read())
        except Exception as e:
            raise DCOSException(
                'Error parsing config file at [{}]: {}'.format(path, e))
        return (MutableToml if mutable else Toml)(toml_obj)


def save(toml_config, config_path=None):
    """
    :param toml_config: TOML configuration object
    :type toml_config: MutableToml or Toml
    :param config_path: path to config to use
    :type config_path: str
    """

    serial = toml.dumps(toml_config._dictionary)
    if config_path is None:
        config_path = get_config_path()

    util.ensure_file_exists(config_path)
    util.enforce_file_permissions(config_path)
    with util.open_file(config_path, 'w') as config_file:
        config_file.write(serial)


def _get_path(toml_config, path):
    """
    :param config: Dict with the configuration values
    :type config: dict
    :param path: Path to the value. E.g. 'path.to.value'
    :type path: str
    :returns: Value stored at the given path
    :rtype: double, int, str, list or dict
    """

    for section in path.split('.'):
        toml_config = toml_config[section]

    return toml_config


def unset(name):
    """
    :param name: name of config value to unset
    :type name: str
    :returns: message of property removed
    :rtype: str
    """

    toml_config = get_config(True)
    toml_config_pre = copy.deepcopy(toml_config)
    section = name.split(".", 1)[0]
    if section not in toml_config_pre._dictionary:
        toml_config_pre._dictionary[section] = {}
    value = toml_config.pop(name, None)

    if value is None:
        raise DCOSException("Property {!r} doesn't exist".format(name))
    elif isinstance(value, collections.Mapping):
        raise DCOSException(_generate_choice_msg(name, value))
    else:
        msg = "Removed [{}]".format(name)
        # dcos_acs_token is coupled to a specific dcos_url
        if name == "core.dcos_url":
            unset_token = bool(toml_config.pop("core.dcos_acs_token", None))
            if unset_token:
                msg += " and [core.dcos_acs_token]"
        save(toml_config)
        return msg


def _generate_choice_msg(name, value):
    """
    :param name: name of the property
    :type name: str
    :param value: dictionary for the value
    :type value: dcos.config.Toml
    :returns: an error message for top level properties
    :rtype: str
    """

    message = ("Property {!r} doesn't fully specify a value - "
               "possible properties are:").format(name)
    for key, _ in sorted(value.property_items()):
        message += '\n{}.{}'.format(name, key)

    return message


def _iterator(parent, dictionary):
    """
    :param parent: Path to the value parameter
    :type parent: str
    :param dictionary: Value of the key
    :type dictionary: collection.Mapping
    :returns: An iterator of tuples for each property and value
    :rtype: iterator of (str, any) where any can be str, int, double, list
    """

    for key, value in dictionary.items():

        new_key = key
        if parent is not None:
            new_key = "{}.{}".format(parent, key)

        if not isinstance(value, collections.Mapping):
            yield (new_key, value)
        else:
            for x in _iterator(new_key, value):
                yield x


def split_key(name):
    """
    :param name: the full property path - e.g. marathon.url
    :type name: str
    :returns: the section and property name
    :rtype: (str, str)
    """

    terms = name.split('.', 1)
    if len(terms) != 2:
        raise DCOSException('Property name must have both a section and '
                            'key: <section>.<key> - E.g. marathon.url')

    return (terms[0], terms[1])


def get_config_schema(command):
    """
    :param command: the subcommand name
    :type command: str
    :returns: the subcommand's configuration schema
    :rtype: dict
    """

    # import here to avoid circular import
    from dcos.subcommand import (
            command_executables, config_schema, default_subcommands)

    # core.* config variables are special.  They're valid, but don't
    # correspond to any particular subcommand, so we must handle them
    # separately.
    if command == "core":
        return json.loads(
            pkg_resources.resource_string(
                'dcos',
                'data/config-schema/core.json').decode('utf-8'))
    elif command in default_subcommands():
        return json.loads(
            pkg_resources.resource_string(
                'dcos',
                'data/config-schema/{}.json'.format(command)).decode('utf-8'))
    else:
        executable = command_executables(command)
        return config_schema(executable, command)


def get_property_description(section, subkey):
    """
    :param section: section of config paramater
    :type section: str
    :param subkey: property within 'section'
    :type subkey: str
    :returns: description of section.subkey or None if no description
    :rtype: str | None
    """

    schema = get_config_schema(section)
    property_info = schema["properties"].get(subkey)
    if property_info is not None:
        return property_info.get("description")
    else:
        raise DCOSException(
            "No schema found found for {}.{}".format(section, subkey))


def check_config(toml_config_pre, toml_config_post, section):
    """
    :param toml_config_pre: dictionary for the value before change
    :type toml_config_pre: dcos.api.config.Toml
    :param toml_config_post: dictionary for the value with change
    :type toml_config_post: dcos.api.config.Toml
    :param section: section of the config to check
    :type section: str
    :returns: process status
    :rtype: int
    """

    errors_pre = util.validate_json(toml_config_pre._dictionary[section],
                                    get_config_schema(section))
    errors_post = util.validate_json(toml_config_post._dictionary[section],
                                     get_config_schema(section))

    logger.info('Comparing changes in the configuration...')
    logger.info('Errors before the config command: %r', errors_pre)
    logger.info('Errors after the config command: %r', errors_post)

    if len(errors_post) != 0:
        if len(errors_pre) == 0:
            raise DCOSException(util.list_to_err(errors_post))

        def _errs(errs):
            return set([e.split('\n')[0] for e in errs])

        diff_errors = _errs(errors_post) - _errs(errors_pre)
        if len(diff_errors) != 0:
            raise DCOSException(util.list_to_err(errors_post))


def generate_choice_msg(name, value):
    """
    :param name: name of the property
    :type name: str
    :param value: dictionary for the value
    :type value: dcos.config.Toml
    :returns: an error message for top level properties
    :rtype: str
    """

    message = ("Property {!r} doesn't fully specify a value - "
               "possible properties are:").format(name)
    for key, _ in sorted(value.property_items()):
        message += '\n{}.{}'.format(name, key)

    return message


def generate_root_schema(toml_config):
    """
    :param toml_configs: dictionary of values
    :type toml_config: TomlConfig
    :returns: configuration_schema
    :rtype: jsonschema
    """

    root_schema = {
        '$schema': 'http://json-schema.org/schema#',
        'type': 'object',
        'properties': {},
        'additionalProperties': False,
    }

    # Load the config schema from all the subsections into the root schema
    for section in toml_config.keys():
        config_schema = get_config_schema(section)
        root_schema['properties'][section] = config_schema

    return root_schema


class Toml(collections.Mapping):
    """Class for getting value from TOML.

    :param dictionary: configuration dictionary
    :type dictionary: dict
    """

    def __init__(self, dictionary):
        self._dictionary = dictionary

    def __getitem__(self, path):
        """
        :param path: Path to the value. E.g. 'path.to.value'
        :type path: str
        :returns: Value stored at the given path
        :rtype: double, int, str, list or dict
        """

        toml_config = _get_path(self._dictionary, path)
        if isinstance(toml_config, collections.Mapping):
            return Toml(toml_config)
        else:
            return toml_config

    def __iter__(self):
        """
        :returns: Dictionary iterator
        :rtype: iterator
        """

        return iter(self._dictionary)

    def property_items(self):
        """Iterator for full-path keys and values

        :returns: Iterator for pull-path keys and values
        :rtype: iterator of tuples
        """

        return _iterator(None, self._dictionary)

    def __len__(self):
        """
        :returns: The length of the dictionary
        :rtype: int
        """

        return len(self._dictionary)


class MutableToml(collections.MutableMapping):
    """Class for managing CLI configuration through TOML.

    :param dictionary: configuration dictionary
    :type dictionary: dict
    """

    def __init__(self, dictionary):
        self._dictionary = dictionary

    def __getitem__(self, path):
        """
        :param path: Path to the value. E.g. 'path.to.value'
        :type path: str
        :returns: Value stored at the given path
        :rtype: double, int, str, list or dict
        """

        toml_config = _get_path(self._dictionary, path)
        if isinstance(toml_config, collections.MutableMapping):
            return MutableToml(toml_config)
        else:
            return toml_config

    def __iter__(self):
        """
        :returns: Dictionary iterator
        :rtype: iterator
        """

        return iter(self._dictionary)

    def property_items(self):
        """Iterator for full-path keys and values

        :returns: Iterator for pull-path keys and values
        :rtype: iterator of tuples
        """

        return _iterator(None, self._dictionary)

    def __len__(self):
        """
        :returns: The length of the dictionary
        :rtype: int
        """

        return len(self._dictionary)

    def __setitem__(self, path, value):
        """
        :param path: Path to set
        :type path: str
        :param value: Value to store
        :type value: double, int, str, list or dict
        """

        toml_config = self._dictionary

        sections = path.split('.')
        for section in sections[:-1]:
            toml_config = toml_config.setdefault(section, {})

        toml_config[sections[-1]] = value

    def __delitem__(self, path):
        """
        :param path: Path to delete
        :type path: str
        """
        toml_config = self._dictionary

        sections = path.split('.')
        for section in sections[:-1]:
            toml_config = toml_config[section]

        del toml_config[sections[-1]]



import pytest
from mock import MagicMock, patch

from dcos import mesos
from dcos.errors import DCOSException
from dcoscli.log import log_files
from dcoscli.task.main import _dcos_log, _metrics, main

from .common import assert_mock


@patch('dcos.config.get_config')
def test_log_master_unavailable(config_mock):
    config_mock.return_value = {'core.dcos_url': 'foo'}

    """ Test master's state.json being unavailable """
    client = mesos.DCOSClient()
    client.get_master_state = _mock_exception()

    with patch('dcos.mesos.DCOSClient', return_value=client):
        args = ['task', 'log', '_']
        assert_mock(main, args, returncode=1, stderr=(b"exception\n"))


def test_log_no_tasks():
    """ Test slave's state.json being unavailable """
    with patch('dcos.mesos.DCOSClient.get_master_state', return_value={}), \
            patch('dcos.mesos.Master.tasks', return_value={}):

        stderr = b"""No matching tasks. Exiting.\n"""
        args = ['task', 'log', 'test-app-1']
        assert_mock(main, args, returncode=1, stderr=stderr)


def test_task_exact_match():
    """Test a task gets returned if it is an exact match, even if
    it is a substring of another task.
    """
    with patch('dcos.mesos.Master.slaves',
               return_value=[{"id": "foo"}, {"id": "foobar"}]):
        master = mesos.Master(None)
        assert master.slave("foo") == {"id": "foo"}


def test_log_file_unavailable():
    """ Test a file's read.json being unavailable """
    files = [mesos.MesosFile('bogus')]
    files[0].read = _mock_exception('exception')

    with pytest.raises(DCOSException) as e:
        log_files(files, True, 10)

    msg = "No files exist. Exiting."
    assert e.exconly().split(':', 1)[1].strip() == msg


def _mock_exception(contents='exception'):
    return MagicMock(side_effect=DCOSException(contents))


@patch('dcos.http.get')
@patch('dcos.config.get_config_val')
def test_dcos_log(mocked_get_config_val, mocked_http_get):
    mocked_get_config_val.return_value = 'http://127.0.0.1'

    m = MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    executor_info = {
        'tasks': [
            {
                'container': 'container1',
                'state': 'TASK_RUNNING',
                'statuses': [
                    {
                        'state': 'TASK_RUNNING',
                        'container_status': {
                            'container_id': {
                                'value': 'child-123',
                                'parent': {
                                    'value': 'parent-456'
                                }
                            }
                        }
                    }
                ],
                'slave_id': 'slave-123',
                'framework_id': 'framework-123',
                'id': 'id-123'
            }
        ]
    }

    task = MagicMock
    task.executor = lambda: executor_info
    _dcos_log(False, [task], 10, 'stdout', False)
    mocked_http_get.assert_called_with(
        'http://127.0.0.1/system/v1/agent/slave-123/logs/v1/range/framework/'
        'framework-123/executor/id-123/container/parent-456.child-123'
        '?skip_prev=10&filter=STREAM:STDOUT',
        headers={'Accept': 'text/plain'})


@patch('dcoscli.log.follow_logs')
@patch('dcos.http.get')
@patch('dcos.config.get_config_val')
def test_dcos_log_stream(mocked_get_config_val, mocked_http_get,
                         mocked_follow_logs):
    mocked_get_config_val.return_value = 'http://127.0.0.1'

    m = MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    executor_info = {
        'tasks': [
            {
                'container': 'container1',
                'state': 'TASK_RUNNING',
                'statuses': [
                    {
                        'state': 'TASK_RUNNING',
                        'container_status': {
                            'container_id': {
                                'value': 'child-123',
                            }
                        }
                    }
                ],
                'slave_id': 'slave-123',
                'framework_id': 'framework-123',
                'id': 'id-123'
            }
        ]
    }

    task = MagicMock
    task.executor = lambda: executor_info
    _dcos_log(True, [task], 10, 'stderr', False)
    mocked_follow_logs.assert_called_with(
        'http://127.0.0.1/system/v1/agent/slave-123/logs/v1/'
        'stream/framework/framework-123/executor/id-123/container/'
        'child-123?skip_prev=10&filter=STREAM:STDERR')


@patch('dcos.http.get')
@patch('dcos.mesos.get_master')
@patch('dcos.config.get_config_val')
def test_dcos_task_metrics_agent_details(mocked_get_config_val,
                                         mocked_get_master,
                                         mocked_http_get):
    mocked_get_config_val.return_value = 'http://127.0.0.1'

    mock_http_response = MagicMock()
    mock_http_response.status_code = 200
    mocked_http_get.return_value = mock_http_response

    mock_master = MagicMock()
    mock_master.task = lambda _: {'slave_id': 'slave_id'}
    mock_master.get_container_id = lambda _: {
        'parent': {},
        'value': 'container_id'
    }
    mocked_get_master.return_value = mock_master

    _metrics(True, 'task_id', False)

    # Metrics should query both endpoints
    mocked_http_get.assert_any_call('http://127.0.0.1/system/v1/agent/'
                                    'slave_id/metrics/v0/containers/'
                                    'container_id/app')
    mocked_http_get.assert_any_call('http://127.0.0.1/system/v1/agent/'
                                    'slave_id/metrics/v0/containers/'
                                    'container_id')


@patch('dcos.http.get')
@patch('dcos.mesos.get_master')
@patch('dcos.config.get_config_val')
def test_dcos_task_metrics_agent_missing_slave(mocked_get_config_val,
                                               mocked_get_master,
                                               mocked_http_get):
    mocked_get_config_val.return_value = 'http://127.0.0.1'

    mock_http_response = MagicMock()
    mock_http_response.status_code = 200
    mocked_http_get.return_value = mock_http_response

    mock_master = MagicMock()
    mock_master.task = lambda _: {}
    mocked_get_master.return_value = mock_master

    # Should a task not have a slave ID, expect an error
    with pytest.raises(DCOSException):
        _metrics(True, 'task_id', False)



import sys

from functools import wraps

import docopt

from dcos import emitting

emitter = emitting.FlatEmitter()


def decorate_docopt_usage(func):
    """Handle DocoptExit exception

    :param func: function
    :type func: function
    :return: wrapped function
    :rtype: function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except docopt.DocoptExit as e:
            emitter.publish("Command not recognized\n")
            emitter.publish(e)
            return 1
        return result
    return wrapper


def confirm(prompt, yes):
    """
    :param prompt: message to display to the terminal
    :type prompt: str
    :param yes: whether to assume that the user responded with yes
    :type yes: bool
    :returns: True if the user responded with yes; False otherwise
    :rtype: bool
    """

    if yes:
        return True
    else:
        count = 0
        while count < 3:
            sys.stdout.write('{} [yes/no] '.format(prompt))
            sys.stdout.flush()
            response = _read_response().lower()
            if response == 'yes' or response == 'y':
                return True
            elif response == 'no' or response == 'n':
                return False
            else:
                msg = "'{}' is not a valid response.".format(response)
                emitter.publish(msg)
                count += 1
        return False


def confirm_text(prompt, confirmation_text):
    """
    :param prompt: message to display to the terminal
    :type prompt: str
    :param expected_text: The text to compare user input to.
    :type expected_text: str
    :returns: True if the user replies with the expected text; False otherwise
    :rtype: bool
    """

    count = 0
    while count < 3:
        sys.stdout.write('{}: '.format(prompt))
        sys.stdout.flush()
        response = _read_response()
        if response == confirmation_text:
            return True
        else:
            msg = "Expected '{}'. You supplied '{}'.".format(
                confirmation_text,
                response
            )
            emitter.publish(msg)
            count += 1
    return False


def _read_response():
    """
    :returns: The content of STDIN
    :rtype: str
    """
    return sys.stdin.readline().strip()



import json
import re
import time

import pytest

from .helpers.common import assert_command, exec_command
from .helpers.marathon import (add_pod, pod, pod_spec_json, pods, remove_pod,
                               watch_all_deployments)
from ..fixtures.marathon import (DOUBLE_POD_FILE_PATH, DOUBLE_POD_ID,
                                 GOOD_POD_FILE_PATH, GOOD_POD_ID,
                                 POD_KILL_FILE_PATH, POD_KILL_ID,
                                 pod_list_fixture, TRIPLE_POD_FILE_PATH,
                                 TRIPLE_POD_ID, UNGOOD_POD_FILE_PATH,
                                 UPDATED_GOOD_POD_FILE_PATH)

_POD_BASE_CMD = ['dcos', 'marathon', 'pod']
_POD_ADD_CMD = _POD_BASE_CMD + ['add']
_POD_KILL_CMD = _POD_BASE_CMD + ['kill']
_POD_LIST_CMD = _POD_BASE_CMD + ['list']
_POD_REMOVE_CMD = _POD_BASE_CMD + ['remove']
_POD_SHOW_CMD = _POD_BASE_CMD + ['show']
_POD_UPDATE_CMD = _POD_BASE_CMD + ['update']


def test_pod_add_from_file():
    add_pod(GOOD_POD_FILE_PATH)
    remove_pod(GOOD_POD_ID, force=False)
    watch_all_deployments()


def test_pod_list():
    expected_json = pod_list_fixture()

    with pods({GOOD_POD_FILE_PATH: GOOD_POD_ID,
               DOUBLE_POD_FILE_PATH: DOUBLE_POD_ID,
               TRIPLE_POD_FILE_PATH: TRIPLE_POD_ID}):

        _assert_pod_list_json(expected_json)
        _assert_pod_list_table()


def test_pod_update_does_not_support_properties():
    cmd = _POD_UPDATE_CMD + ['any-pod', 'foo=bar']
    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 1
    assert stdout.startswith(b'Command not recognized\n')
    assert stderr == b''


def test_pod_update_from_stdin():
    with pod(GOOD_POD_FILE_PATH, GOOD_POD_ID):
        # The deployment will never complete
        _assert_pod_update_from_stdin(
            extra_args=[],
            pod_json_file_path=UNGOOD_POD_FILE_PATH)

        # Override above failed deployment
        _assert_pod_update_from_stdin(
            extra_args=['--force'],
            pod_json_file_path=UPDATED_GOOD_POD_FILE_PATH)

        watch_all_deployments()


@pytest.mark.skipif(
    True, reason='https://mesosphere.atlassian.net/browse/DCOS-13368')
def test_pod_kill():
    with pod(POD_KILL_FILE_PATH, POD_KILL_ID):
        kill_1, keep, kill_2 = _get_pod_instance_ids(POD_KILL_ID, 3)

        remove_args = [POD_KILL_ID, kill_1, kill_2]
        assert_command(_POD_KILL_CMD + remove_args)

        new_instance_ids = _get_pod_instance_ids(POD_KILL_ID, 3)
        assert keep in new_instance_ids
        assert kill_1 not in new_instance_ids
        assert kill_2 not in new_instance_ids
        # Marathon spins up new instances to replace the killed ones
        assert len(new_instance_ids) == 3


def _pod_add_from_stdin(file_path):
    cmd = _POD_ADD_CMD
    with open(file_path) as fd:
        returncode, stdout, stderr = exec_command(cmd, stdin=fd)

    assert returncode == 0
    assert re.fullmatch('Created deployment \S+\n', stdout.decode('utf-8'))
    assert stderr == b''

    watch_all_deployments()


def _assert_pod_list_json(expected_json):
    cmd = _POD_LIST_CMD + ['--json']
    returncode, stdout, stderr = exec_command(cmd)
    assert returncode == 0
    assert stderr == b''

    actual_json = json.loads(stdout.decode('utf-8'))
    _assert_pod_list_json_subset(expected_json, actual_json)


def _assert_pod_list_json_subset(expected_json, actual_json):
    actual_pods_by_id = {pod['id']: pod for pod in actual_json}

    for expected_pod in expected_json:
        pod_id = expected_pod['id']
        actual_pod = actual_pods_by_id[pod_id]
        pod_spec_json(expected_pod['spec'], actual_pod['spec'])

    assert len(actual_json) == len(expected_json)


def _assert_pod_list_table():
    _wait_for_instances({'/double-pod': 2, '/good-pod': 1, '/winston': 1})
    returncode, stdout, stderr = exec_command(_POD_LIST_CMD)

    assert returncode == 0
    assert stderr == b''

    stdout_lines = stdout.decode('utf-8').split('\n')

    pattern = r'ID\+TASKS +INSTANCES +VERSION +STATUS +STATUS SINCE +WAITING *'
    assert re.fullmatch(pattern, stdout_lines[0])

    assert stdout_lines[1].startswith('/double-pod')
    assert stdout_lines[2].startswith(' |-thing-1')
    assert stdout_lines[3].startswith(' |-thing-2')
    assert stdout_lines[4].startswith('/good-pod')
    assert stdout_lines[5].startswith(' |-good-container')
    assert stdout_lines[6].startswith('/winston')
    assert stdout_lines[7].startswith(' |-the-cat')
    assert stdout_lines[8].startswith(' |-thing-1')
    assert stdout_lines[9].startswith(' |-thing-2')

    assert stdout_lines[10] == ''
    assert len(stdout_lines) == 11


def _assert_pod_update_from_stdin(extra_args, pod_json_file_path):
    cmd = _POD_UPDATE_CMD + [GOOD_POD_ID] + extra_args
    with open(pod_json_file_path) as fd:
        returncode, stdout, stderr = exec_command(cmd, stdin=fd)

    assert returncode == 0
    assert re.fullmatch('Created deployment \S+\n', stdout.decode('utf-8'))
    assert stderr == b''


def _wait_for_instances(expected_instances, max_attempts=10):
    """Polls the `pod list` command until the instance counts are as expected.

    :param expected_instances: a mapping from pod ID to instance count
    :type expected_instances: {}
    :param max_attempts: give up and fail the test after this many attempts
    :type max_attempts: int
    :rtype: None
    """

    for attempt in range(max_attempts):
        returncode, stdout, stderr = exec_command(_POD_LIST_CMD + ['--json'])

        assert returncode == 0
        assert stderr == b''

        status_json = json.loads(stdout.decode('utf-8'))
        actual_instances = {pod_obj['id']: len(pod_obj.get('instances', []))
                            for pod_obj in status_json}

        if actual_instances == expected_instances:
            return
        time.sleep(1)
    else:
        assert False, "Timed out waiting for expected instance counts"


def _get_pod_instance_ids(pod_id, target_instance_count):
    """Waits for the given pod to reach a target instance count, then returns
    the IDs of all instances.

    :param pod_id: the pod to retrieve the instance IDs from
    :type pod_id: str
    :param target_instance_count: waits until the number of instances reaches
                                  this number
    :type target_instance_count: int
    :returns: a tuple of the pod's instance IDs
    :rtype: tuple(str)
    """

    _wait_for_instances({'/{}'.format(pod_id): target_instance_count})

    show_cmd = _POD_SHOW_CMD + [pod_id]
    returncode, stdout, stderr = exec_command(show_cmd)

    assert returncode == 0
    assert stderr == b''

    pod_status_json = json.loads(stdout.decode('utf-8'))
    instances_json = pod_status_json['instances']
    return tuple(instance['id'] for instance in instances_json)



import contextlib
import os

import pytest

from dcos import constants, util
from dcos.errors import DCOSException


def test_open_file():
    path = 'nonexistant_file_name.txt'
    with pytest.raises(DCOSException) as excinfo:
        with util.open_file(path):
            pass
    assert 'Error opening file [{}]: No such file or directory'.format(path) \
        in str(excinfo.value)


@contextlib.contextmanager
def env():
    """Context manager for altering env vars in tests """

    try:
        old_env = dict(os.environ)
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)


def add_cluster_dir(cluster_id, dcos_dir):
    clusters_dir = os.path.join(dcos_dir, constants.DCOS_CLUSTERS_SUBDIR)
    util.ensure_dir_exists(clusters_dir)

    cluster_path = os.path.join(clusters_dir, cluster_id)
    util.ensure_dir_exists(cluster_path)

    os.path.join(cluster_path, "dcos.toml")
    return cluster_path


def create_global_config(dcos_dir):
    global_toml = os.path.join(dcos_dir, "dcos.toml")
    util.ensure_file_exists(global_toml)
    return global_toml



import collections

import docopt

import dcoscli
from dcos import cmds, config, emitting, http, util
from dcos.errors import DCOSException, DefaultError
from dcoscli.cluster.main import setup
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import decorate_docopt_usage

emitter = emitting.FlatEmitter()
logger = util.get_logger(__name__)


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("config"),
        argv=argv,
        version='dcos-config version {}'.format(dcoscli.version))

    http.silence_requests_warnings()

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: all the supported commands
    :rtype: list of dcos.cmds.Command
    """

    return [
        cmds.Command(
            hierarchy=['config', 'set'],
            arg_keys=['<name>', '<value>'],
            function=_set),

        cmds.Command(
            hierarchy=['config', 'unset'],
            arg_keys=['<name>'],
            function=_unset),

        cmds.Command(
            hierarchy=['config', 'show'],
            arg_keys=['<name>'],
            function=_show),

        cmds.Command(
            hierarchy=['config', 'validate'],
            arg_keys=[],
            function=_validate),

        cmds.Command(
            hierarchy=['config'],
            arg_keys=['--info'],
            function=_info),
    ]


def _info(info):
    """
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: process status
    :rtype: int
    """

    emitter.publish(default_command_info("config"))
    return 0


def _set(name, value):
    """
    :returns: process status
    :rtype: int
    """

    if name == "core.dcos_url":
        return _cluster_setup(value)

    toml, msg = config.set_val(name, value)
    emitter.publish(DefaultError(msg))

    return 0


def _cluster_setup(dcos_url):
    """
    Setup a cluster using "cluster" directory instead "global" directory, until
    we deprecate "global" config command: `dcos config set core.dcos_url x`
    """

    notice = ("This config property is being deprecated. "
              "To setup the CLI to talk to your cluster, please run "
              "`dcos cluster setup <dcos_url>`.")
    emitter.publish(DefaultError(notice))

    return setup(dcos_url)


def _unset(name):
    """
    :returns: process status
    :rtype: int
    """

    msg = config.unset(name)
    emitter.publish(DefaultError(msg))

    return 0


def _format_config(file_value, effective_value, name=None, envvar_name=None):
    """
    Construct a string to show on a terminal, indicating the value and
    possibly other useful things such as the setting name and whether
    it is being controlled by an environment variable.

    >>> _format_config('x', 'x')
    'x'
    >>> _format_config('x', 'x', 'setting.name') ->
    'setting.name x'
    >>> _format_config('x', 'y', envvar_name='ENVVAR')
    'y # overwritten by ENVVAR; config file value: x'
    >>> _format_config('x', 'y', 'setting.name', envvar_name='ENVVAR')
    'setting.name y   # overwritten by ENVVAR; config file value: x'

    :param file_value: config value present in the toml file
    :type file_value: str

    :param effective_value: config value either from file or as overwritten
                            from the environment
    :type effective_value: str
    :param name: config key (not used for single value show)
    :type name: str|None
    :param envvar_name: name of environment variable that overwote the value
    :type envvar_name: str|None
    :returns: formatted string for terminal
    :rtype: str
    """

    # when declaring that vars are overwritten by the environment,
    # line up those messages to this column (unless the var name is long)
    overwite_msg_display_column = 35

    # When showing all values, don't print the token value;
    if name == "core.dcos_acs_token":
        print_value = "*"*8
    else:
        print_value = effective_value

    if file_value == effective_value:
        if name:
            return '%s %s' % (name, print_value)
        else:
            return effective_value
    else:
        if not envvar_name:
            envvar_name = "N/A"  # this should never happen
        if name:
            s = '%s %s' % (name, print_value)
        else:
            s = effective_value

        left_pad_fmt = '%-{}s'.format(overwite_msg_display_column)  # '%-35s'

        msg_start = left_pad_fmt + ' # overwritten by environment var %s; '

        if print_value != effective_value:
            # We're obscuring the effective security token
            # so don't report the file value either
            msg = msg_start + "config file value differs"
            return msg % (s, envvar_name)

        msg = msg_start + 'config file value: %s'
        return msg % (s, envvar_name, file_value)


def _show(name):
    """
    :returns: process status
    :rtype: int
    """

    toml_config = config.get_config(True)

    if name is not None:
        file_value = toml_config.get(name)
        try:
            # If the user presented a partial key name, eg 'core' when
            # we have 'core.xyz'; we will get an exception here
            effective_value, envvar_name = config.get_config_val_envvar(name)
        except DCOSException as e:
            # The expected case of a partial key name has special
            # handling via this mechanism.
            if isinstance(file_value, collections.Mapping):
                exc_msg = config.generate_choice_msg(name, file_value)
                raise DCOSException(exc_msg)
            raise  # Unexpected errors, pass right along

        if effective_value is None:
            raise DCOSException("Property {!r} doesn't exist".format(name))
        else:
            msg = _format_config(file_value, effective_value,
                                 envvar_name=envvar_name)
            emitter.publish(msg)

    else:
        # Let's list all of the values
        for key, value in sorted(toml_config.property_items()):
            file_value = toml_config.get(key)
            effective_value, envvar_name = config.get_config_val_envvar(key)

            msg = _format_config(file_value, effective_value, key,
                                 envvar_name=envvar_name)
            emitter.publish(msg)

    return 0


def _validate():
    """
    :returns: process status
    :rtype: int
    """

    toml_config = config.get_config(True)

    emitter.publish('Validating %s ...' % config.get_config_path())
    errs = util.validate_json(toml_config._dictionary,
                              config.generate_root_schema(toml_config))
    if len(errs) != 0:
        emitter.publish(util.list_to_err(errs))
        return 1

    emitter.publish("Congratulations, your configuration is valid!")
    return 0



import copy
import datetime
import operator
import posixpath

import textwrap

from collections import OrderedDict

import prettytable

from dcos import auth, marathon, mesos, util

EMPTY_ENTRY = '---'

DEPLOYMENT_DISPLAY = {'ResolveArtifacts': 'artifacts',
                      'ScaleApplication': 'scale',
                      'StartApplication': 'start',
                      'StopApplication': 'stop',
                      'RestartApplication': 'restart',
                      'ScalePod': 'scale',
                      'StartPod': 'start',
                      'StopPod': 'stop',
                      'RestartPod': 'restart',
                      'KillAllOldTasksOf': 'kill-tasks'}

logger = util.get_logger(__name__)


def task_table(tasks):
    """Returns a PrettyTable representation of the provided mesos tasks.

    :param tasks: tasks to render
    :type tasks: [Task]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ("NAME", lambda t: t["name"]),
        ("HOST", lambda t: t.slave()["hostname"]),
        ("USER", lambda t: t.user()),
        ("STATE", lambda t: t["state"].split("_")[-1][0]),
        ("ID", lambda t: t["id"]),
        ("MESOS ID", lambda t: t["slave_id"]),
    ])

    tb = table(fields, tasks, sortby="NAME")
    tb.align["NAME"] = "l"
    tb.align["HOST"] = "l"
    tb.align["ID"] = "l"
    tb.align["MESOS ID"] = "l"

    return tb


def app_table(apps, deployments):
    """Returns a PrettyTable representation of the provided apps.

    :param apps: apps to render
    :type apps: [dict]
    :param deployments: deployments to enhance information
    :type deployments: [dict]
    :rtype: PrettyTable
    """

    deployment_map = {}
    for deployment in deployments:
        deployment_map[deployment['id']] = deployment

    def get_cmd(app):
        if app["cmd"] is not None:
            return app["cmd"]
        else:
            return app["args"]

    def get_container(app):
        if app["container"] is not None:
            return app["container"]["type"]
        else:
            return "mesos"

    def get_health(app):
        if app["healthChecks"]:
            return "{}/{}".format(app["tasksHealthy"],
                                  app["tasksRunning"])
        else:
            return EMPTY_ENTRY

    def get_deployment(app):
        deployment_ids = {deployment['id']
                          for deployment in app['deployments']}

        actions = []
        for deployment_id in deployment_ids:
            deployment = deployment_map.get(deployment_id)
            if deployment:
                for action in deployment['currentActions']:
                    if action['app'] == app['id']:
                        actions.append(DEPLOYMENT_DISPLAY[action['action']])

        if len(actions) == 0:
            return EMPTY_ENTRY
        elif len(actions) == 1:
            return actions[0]
        else:
            return "({})".format(", ".join(actions))

    fields = OrderedDict([
        ("ID", lambda a: a["id"]),
        ("MEM", lambda a: a["mem"]),
        ("CPUS", lambda a: a["cpus"]),
        ("TASKS", lambda a: "{}/{}".format(a["tasksRunning"],
                                           a["instances"])),
        ("HEALTH", get_health),
        ("DEPLOYMENT", get_deployment),
        ("WAITING", lambda app: app.get('overdue', False)),
        ("CONTAINER", get_container),
        ("CMD", get_cmd)
    ])

    limits = {
        "CMD": 35
    }

    tb = truncate_table(fields, apps, limits, sortby="ID")
    tb.align["CMD"] = "l"
    tb.align["ID"] = "l"
    tb.align["WAITING"] = "l"

    return tb


def app_task_table(tasks):
    """Returns a PrettyTable representation of the provided marathon tasks.

    :param tasks: tasks to render
    :type tasks: [dict]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ("APP", lambda t: t["appId"]),
        ("HEALTHY", lambda t:
         all(check['alive'] for check in t.get('healthCheckResults', []))),
        ("STARTED", lambda t: t.get("startedAt", "N/A")),
        ("HOST", lambda t: t["host"]),
        ("ID", lambda t: t["id"])
    ])

    tb = table(fields, tasks, sortby="APP")
    tb.align["APP"] = "l"
    tb.align["ID"] = "l"

    return tb


def deployment_table(deployments):
    """Returns a PrettyTable representation of the provided marathon
    deployments.

    :param deployments: deployments to render
    :type deployments: [dict]
    :rtype: PrettyTable

    """

    def join_path_ids(deployment, affected_resources_key):
        """Create table cell for "affectedApps"/"affectedPods" in deployment.

        :param deployment: the deployment JSON to read
        :type deployment: {}
        :param affected_resources_key: either "affectedApps" or "affectedPods"
        :type affected_resources_key: str
        :returns: newline-separated path IDs if they exist, otherwise an empty
                  cell indicator
        :rtype: str
        """

        path_ids = deployment.get(affected_resources_key)
        return '\n'.join(path_ids) if path_ids else '-'

    def resource_path_id(action):
        """Get the path ID of the app or pod represented by the given action.

        :param action: the Marathon deployment action JSON object to read
        :type action: {}
        :returns: the value of the "app" or "pod" field if it exists, else None
        :rtype: str
        """

        path_id = action.get('app') or action.get('pod')

        if path_id is None:
            template = 'Expected "app" or "pod" field in action: %s'
            logger.exception(template, action)

        return path_id

    def get_action(deployment):

        multiple_resources = len({resource_path_id(action) for action in
                                  deployment['currentActions']}) > 1

        ret = []
        for action in deployment['currentActions']:
            try:
                action_display = DEPLOYMENT_DISPLAY[action['action']]
            except KeyError:
                logger.exception('Missing action entry')

                raise ValueError(
                    'Unknown Marathon action: {}'.format(action['action']))

            if resource_path_id(action) is None:
                ret.append('N/A')
            elif multiple_resources:
                path_id = resource_path_id(action)
                ret.append('{0} {1}'.format(action_display, path_id))
            else:
                ret.append(action_display)

        return '\n'.join(ret)

    fields = OrderedDict([
        ('APP', lambda d: join_path_ids(d, 'affectedApps')),
        ('POD', lambda d: join_path_ids(d, 'affectedPods')),
        ('ACTION', get_action),
        ('PROGRESS', lambda d: '{0}/{1}'.format(d['currentStep']-1,
                                                d['totalSteps'])),
        ('ID', lambda d: d['id'])
    ])

    tb = table(fields, deployments, sortby="APP")
    tb.align['APP'] = 'l'
    tb.align['POD'] = 'l'
    tb.align['ACTION'] = 'l'
    tb.align['ID'] = 'l'

    return tb


def service_table(services):
    """Returns a PrettyTable representation of the provided DC/OS services.

    :param services: services to render
    :type services: [Framework]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ("NAME", lambda s: s['name']),
        ("HOST", lambda s: s['hostname']),
        ("ACTIVE", lambda s: s['active']),
        ("TASKS", lambda s: len(s['tasks'])),
        ("CPU", lambda s: s['resources']['cpus']),
        ("MEM", lambda s: s['resources']['mem']),
        ("DISK", lambda s: s['resources']['disk']),
        ("ID", lambda s: s['id']),
    ])

    tb = table(fields, services, sortby="NAME")
    tb.align["ID"] = 'l'
    tb.align["NAME"] = 'l'

    return tb


def job_table(job_list):
    """Returns a PrettyTable representation of the job list from Metronome.

    :param job_list: jobs to render
    :type job_list: [job]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ('id', lambda s: s['id']),
        ('Status', lambda s: _job_status(s)),
        ('Last Run', lambda s: _last_run_status(s)),
    ])

    tb = truncate_table(fields, job_list, None, sortby="ID")
    tb.align["STATUS"] = 'l'
    tb.align["LAST RUN"] = 'l'

    return tb


def job_history_table(schedule_list):
    """Returns a PrettyTable representation of the job history from Metronome.

    :param schedule_list: job schedule list to render
    :type schedule_list: [history]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ('id', lambda s: s['id']),
        ('started', lambda s: s['createdAt']),
        ('finished', lambda s: s['finishedAt']),
    ])
    tb = table(fields, schedule_list, sortby="STARTED")
    tb.align["STARTED"] = 'l'
    tb.align["FINISHED"] = 'l'

    return tb


def schedule_table(schedule_list):
    """Returns a PrettyTable representation of the schedule list of a job
    from Metronome.

    :param schedule_list: schedules to render
    :type schedule_list: [schedule]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ('id', lambda s: s['id']),
        ('cron', lambda s: s['cron']),
        ('enabled', lambda s: s['enabled']),
        ('concurrency policy', lambda s: s['concurrencyPolicy']),
        ('next run', lambda s: s['nextRunAt']),
    ])
    tb = table(fields, schedule_list)
    tb.align['CRON'] = 'l'
    tb.align['ENABLED'] = 'l'
    tb.align['NEXT RUN'] = 'l'
    tb.align['CONCURRENCY POLICY'] = 'l'

    return tb


def job_runs_table(runs_list):
    """Returns a PrettyTable representation of the runs list of a job from
    Metronome.

    :param runs_list: current runs of a job to render
    :type runs_list: [runs]
    :rtype: PrettyTable
    """
    fields = OrderedDict([
        ('id', lambda s: s['id']),
        ('job id', lambda s: s['jobId']),
        ('started at', lambda s: s['createdAt']),
    ])
    tb = table(fields, runs_list)
    tb.align['JOB ID'] = 'l'
    tb.align['STARTED AT'] = 'l'

    return tb


def _str_to_datetime(datetime_str):
    """ Takes a JSON date of `2017-03-30T15:50:16.187+0000` format and
    Returns a datetime.

    :param datetime_str: JSON date
    :type datetime_str: str
    :rtype: datetime
    """
    if not datetime_str:
        return None
    datetime_str = datetime_str.split('+')[0]
    return datetime.datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%S.%f")


def _last_run_status(job):
    """ Provided a job with embedded history it Returns a status based on the
    following rules:
        0 Runs = 'N/A'
        last success is > last failure = 'Success' otherwise 'Failed'

    :param job: JSON job with embedded history
    :type job: dict
    :rtype: str
    """
    last_success = _str_to_datetime(job['historySummary']['lastSuccessAt'])
    last_failure = _str_to_datetime(job['historySummary']['lastFailureAt'])
    if not last_success and not last_failure:
        return 'N/A'
    elif ((last_success and not last_failure) or
          (last_success and last_success > last_failure)):
        return 'Success'
    else:
        return 'Failed'


def _job_status(job):
    """Utility function that returns the status of a job

    :param job: job json
    :type job: json
    :rtype: str

    """

    if 'activeRuns' in job:
        return "Running"
    # short circuit will prevent failure
    elif 'schedules' not in job or not job['schedules']:
        return "Unscheduled"
    else:
        return "Scheduled"


def _count_apps(group, group_dict):
    """Counts how many apps are registered for each group.  Recursively
    populates the profided `group_dict`, which maps group_id ->
    (group, count).

    :param group: nested group dictionary
    :type group: dict
    :param group_dict: group map that maps group_id -> (group, count)
    :type group_dict: dict
    :rtype: dict

    """

    for child_group in group['groups']:
        _count_apps(child_group, group_dict)

    count = (len(group['apps']) +
             sum(group_dict[child_group['id']][1]
                 for child_group in group['groups']))

    group_dict[group['id']] = (group, count)


def group_table(groups):
    """Returns a PrettyTable representation of the provided marathon
    groups

    :param groups: groups to render
    :type groups: [dict]
    :rtype: PrettyTable

    """

    group_dict = {}
    for group in groups:
        _count_apps(group, group_dict)

    fields = OrderedDict([
        ('ID', lambda g: g[0]['id']),
        ('APPS', lambda g: g[1]),
    ])

    tb = table(fields, group_dict.values(), sortby="ID")
    tb.align['ID'] = 'l'

    return tb


def pod_table(pods):
    """Returns a PrettyTable representation of the provided Marathon pods.

    :param pods: pods to render
    :type pods: [dict]
    :rtype: PrettyTable
    """

    def id_and_containers(pod):
        """Extract the pod ID and container names from the given pod JSON.

        :param pod: the pod JSON to read
        :type pod: {}
        :returns: the entry for the ID+CONTAINER column of the pod table
        :rtype: str
        """

        pod_id = pod['id']
        container_names = sorted(container['name'] for container
                                 in pod['spec']['containers'])

        container_lines = ('\n |-{}'.format(name) for name in container_names)
        return pod_id + ''.join(container_lines)

    key_column = 'ID+TASKS'
    fields = OrderedDict([
        (key_column, id_and_containers),
        ('INSTANCES', lambda pod: len(pod.get('instances', []))),
        ('VERSION', lambda pod: pod['spec'].get('version', '-')),
        ('STATUS', lambda pod: pod['status']),
        ('STATUS SINCE', lambda pod: pod['statusSince']),
        ('WAITING', lambda pod: pod.get('overdue', False))
    ])

    tb = table(fields, pods, sortby=key_column)
    tb.align[key_column] = 'l'
    tb.align['VERSION'] = 'l'
    tb.align['STATUS'] = 'l'
    tb.align['STATUS SINCE'] = 'l'
    tb.align['WAITING'] = 'l'

    return tb


def queued_apps_table(queued_apps):
    """Returns a PrettyTable representation of the Marathon
    launch queue content.

    :param queued_apps: apps to render
    :type queued_apps: [dict]
    :rtype: PrettyTable
    """

    def extract_value_from_entry(entry, value):
        """Extracts the value parameter from given row entry. If value
        is not present, EMPTY_ENTRY will be returned

        :param entry: row entry
        :type entry: [dict]
        :param value: value which should be extracted
        :type value: string
        :rtype: str
        """
        return entry.get('processedOffersSummary', {}).get(value, EMPTY_ENTRY)

    key_column = 'ID'
    fields = OrderedDict([
        (key_column, lambda entry: marathon.get_app_or_pod_id(entry)),
        ('SINCE', lambda entry:
            entry.get('since', EMPTY_ENTRY)
         ),
        ('INSTANCES TO LAUNCH', lambda entry:
            entry.get('count', EMPTY_ENTRY)
         ),
        ('WAITING', lambda entry:
            entry.get('delay', {}).get('overdue', EMPTY_ENTRY)
         ),
        ('PROCESSED OFFERS', lambda entry:
            extract_value_from_entry(entry, 'processedOffersCount')
         ),
        ('UNUSED OFFERS', lambda entry:
            extract_value_from_entry(entry, 'unusedOffersCount')
         ),
        ('LAST UNUSED OFFER', lambda entry:
            extract_value_from_entry(entry, 'lastUnusedOfferAt')
         ),
        ('LAST USED OFFER', lambda entry:
            extract_value_from_entry(entry, 'lastUsedOfferAt')
         ),
    ])

    tb = table(fields, queued_apps, sortby=key_column)
    tb.align[key_column] = 'l'
    tb.align['SINCE'] = 'l'
    tb.align['INSTANCES TO LAUNCH'] = 'l'
    tb.align['WAITING'] = 'l'
    tb.align['PROCESSED OFFERS'] = 'l'
    tb.align['UNUSED OFFERS'] = 'l'
    tb.align['LAST UNUSED OFFER'] = 'l'
    tb.align['LAST USED OFFER'] = 'l'

    return tb


def queued_app_table(queued_app):
    """Returns a PrettyTable representation of the Marathon
    launch queue content.

    :param queued_app: app to render
    :type queued_app: dict
    :rtype: PrettyTable
    """

    def calc_division(dividend, divisor):
        """Calcs divident / divisor, displays 0 if divisor equals 0.

        :param dividend: divident
        :type dividend: int
        :param divisor: divisor
        :type divisor: int
        :rtype: str
        """
        if divisor == 0:
            return 0
        else:
            return 100 * dividend / divisor

    def add_reason_entry(calculations, key, requested, reason_entry):
        """Pretty prints the division of
        reason_entry.get('declined') / reason_entry.get('processed')

        :param calculations: object where result should be added
        :type calculations: dict
        :param key: key for which the result should be added
        :type key: string
        :param requested: the value initially was requested for this entry
        :type requested: string
        :param reason_entry: entry for a declined offer reason
        :type reason_entry: [dict]
        :rtype: str
        """
        dividend = reason_entry.get('processed', 0) - \
            reason_entry.get('declined', 0)
        divisor = reason_entry.get('processed', 0)
        calculations[key]['REQUESTED'] = requested
        calculations[key]['MATCHED'] = '{0} / {1}'\
            .format(dividend, divisor)
        if divisor > 0:
            calculations[key]['PERCENTAGE'] = '{0:0.2f}%' \
                .format(calc_division(dividend, divisor))
        else:
            calculations[key]['PERCENTAGE'] = EMPTY_ENTRY

    def extract_reason_from_list(list, reason_string):
        """Extracts the reason for the given reason_string from the given list

        :param list: list of reason entries
        :type list: [dict]
        :param reason_string: reasong as string
        :type reason_string: str
        :rtype: reason entry
        """
        filtered = [x for x in list if x['reason'] == reason_string]
        if len(filtered) == 1:
            return filtered[0]
        else:
            return {'reason': reason_string, 'declined': 0, 'processed': 0}

    fields = OrderedDict([
        ('RESOURCE', lambda entry:
            calculations.get(entry, {}).get('RESOURCE', EMPTY_ENTRY)
         ),
        ('REQUESTED', lambda entry:
            calculations.get(entry, {}).get('REQUESTED', EMPTY_ENTRY)
         ),
        ('MATCHED', lambda entry:
            calculations.get(entry, {}).get('MATCHED', EMPTY_ENTRY)
         ),
        ('PERCENTAGE', lambda entry:
            calculations.get(entry, {}).get('PERCENTAGE', EMPTY_ENTRY)
         ),
    ])

    summary = queued_app.get('processedOffersSummary', {})
    reasons = summary.get('rejectSummaryLastOffers', {})

    declined_by_role = extract_reason_from_list(
        reasons, 'UnfulfilledRole')
    declined_by_constraints = extract_reason_from_list(
        reasons, 'UnfulfilledConstraint')
    declined_by_cpus = extract_reason_from_list(
        reasons, 'InsufficientCpus')
    declined_by_mem = extract_reason_from_list(
        reasons, 'InsufficientMemory')
    declined_by_disk = extract_reason_from_list(
        reasons, 'InsufficientDisk')
    """declined_by_gpus = extract_reason_from_list(
        reasons, 'InsufficientGpus')"""
    declined_by_ports = extract_reason_from_list(
        reasons, 'InsufficientPorts')

    app = queued_app.get('app')
    if app:
        roles = app.get('acceptedResourceRoles', [])
        if len(roles) == 0:
            spec_roles = '[*]'
        else:
            spec_roles = roles
        spec_constraints = app.get('constraints', EMPTY_ENTRY)
        spec_cpus = app.get('cpus', EMPTY_ENTRY)
        spec_mem = app.get('mem', EMPTY_ENTRY)
        spec_disk = app.get('disk', EMPTY_ENTRY)
        """spec_gpus = app.get('gpus', EMPTY_ENTRY)"""
        spec_ports = app.get('ports', EMPTY_ENTRY)
    else:
        def sum_resources(value):
            def container_value(container):
                return container.get('resources', {}).get(value, 0)

            """While running pods, marathon will add resources for
            the executor to the requested resources.
            Therefore this requirements should be reflected in the summary."""
            def executor_value():
                return pod.get('executorResources', {}).get(value, 0)

            resources = sum(map(container_value, pod.get('containers', [])))
            return resources + executor_value()

        pod = queued_app.get('pod')
        roles = pod.\
            get('scheduling', {}).get('placement', {}).\
            get('acceptedResourceRoles', [])
        if len(roles) == 0:
            spec_roles = '[*]'
        else:
            spec_roles = roles
        spec_constraints = pod.\
            get('scheduling', {}).get('placement', {}).\
            get('constraints', EMPTY_ENTRY)
        spec_cpus = sum_resources('cpus')
        spec_mem = sum_resources('mem')
        spec_disk = sum_resources('disk')
        """spec_gpus = sum_resources('gpus')"""
        spec_ports = []
        for container in pod.get('containers', []):
            for endpoint in container.get('endpoints', []):
                spec_ports.append(endpoint.get('hostPort'))

    """'GPUS'"""
    rows = ['ROLE', 'CONSTRAINTS', 'CPUS', 'MEM', 'DISK', 'PORTS']

    calculations = {}
    for reason in rows:
        calculations[reason] = {}
        calculations[reason]['RESOURCE'] = reason

    add_reason_entry(calculations, 'ROLE', spec_roles, declined_by_role)
    add_reason_entry(
        calculations, 'CONSTRAINTS', spec_constraints,
        declined_by_constraints)
    add_reason_entry(calculations, 'CPUS', spec_cpus, declined_by_cpus)
    add_reason_entry(calculations, 'MEM', spec_mem, declined_by_mem)
    add_reason_entry(calculations, 'DISK', spec_disk, declined_by_disk)
    """
    add_reason_entry(calculations, 'GPUS', spec_gpus, declined_by_gpus)
    """
    add_reason_entry(calculations, 'PORTS', spec_ports, declined_by_ports)

    tb = table(fields, rows)
    tb.align['RESOURCE'] = 'l'
    tb.align['REQUESTED'] = 'l'
    tb.align['MATCHED'] = 'l'
    tb.align['PERCENTAGE'] = 'l'

    return tb


def queued_app_details_table(queued_app):
    """Returns a PrettyTable representation of the Marathon
    launch queue detailed content.

    :param queued_app: app to render
    :type queued_app: dict
    :rtype: PrettyTable
    """

    def value_declined(entry, value):
        """Returns `yes` if the value was inside entry.get('reason'),
        returns `no` otherwise.

        :param entry: row entry
        :type entry: [dict]
        :param value: value which should be checked
        :type value: string
        :rtype: PrettyTable
        """
        if value not in entry.get('reason', []):
            return 'ok'
        else:
            return '-'

    reasons = queued_app.get('lastUnusedOffers')
    fields = OrderedDict([
        ('HOSTNAME', lambda entry:
            entry.get('offer', {}).get('hostname', EMPTY_ENTRY)
         ),
        ('ROLE', lambda entry: value_declined(entry, 'UnfulfilledRole')),
        ('CONSTRAINTS', lambda entry:
            value_declined(entry, 'UnfulfilledConstraint')
         ),
        ('CPUS', lambda entry: value_declined(entry, 'InsufficientCpus')),
        ('MEM', lambda entry: value_declined(entry, 'InsufficientMemory')),
        ('DISK', lambda entry: value_declined(entry, 'InsufficientDisk')),
        ('PORTS', lambda entry: value_declined(entry, 'UnfulfilledRole')),
        ('RECEIVED', lambda entry:
            entry.get('timestamp', EMPTY_ENTRY)
         ),
    ])
    """('GPUS', lambda entry: value_declined(entry, 'InsufficientGpus')),"""

    tb = table(fields, reasons, sortby='HOSTNAME')
    tb.align['HOSTNAME'] = 'l'
    tb.align['REASON'] = 'l'
    tb.align['RECEIVED'] = 'l'

    return tb


def package_table(packages):
    """Returns a PrettyTable representation of the provided DC/OS packages

    :param packages: packages to render
    :type packages: [dict]
    :rtype: PrettyTable

    """

    fields = OrderedDict([
        ('NAME', lambda p: p['name']),
        ('VERSION', lambda p: p['version']),
        ('APP',
         lambda p: '\n'.join(p['apps']) if p.get('apps') else EMPTY_ENTRY),
        ('COMMAND',
         lambda p: p['command']['name'] if 'command' in p else EMPTY_ENTRY),
        ('DESCRIPTION', lambda p: p['description'])
    ])

    limits = {
        "DESCRIPTION": 65
    }

    tb = truncate_table(fields, packages, limits, sortby="NAME")
    tb.align['NAME'] = 'l'
    tb.align['VERSION'] = 'l'
    tb.align['APP'] = 'l'
    tb.align['COMMAND'] = 'l'
    tb.align['DESCRIPTION'] = 'l'

    return tb


def package_search_table(search_results):
    """Returns a PrettyTable representation of the provided DC/OS package
    search results

    :param search_results: search_results, in the format of
                           dcos.package.IndexEntries::as_dict()
    :type search_results: [dict]
    :rtype: PrettyTable

    """

    fields = OrderedDict([
        ('NAME', lambda p: p['name']),
        ('VERSION', lambda p: p['currentVersion']),
        ('SELECTED', lambda p: p.get("selected", False)),
        ('FRAMEWORK', lambda p: p['framework']),
        ('DESCRIPTION', lambda p: p['description']
            if len(p['description']) < 77 else p['description'][0:77] + "...")
    ])

    packages = []
    for package in search_results['packages']:
        package_ = copy.deepcopy(package)
        packages.append(package_)

    tb = table(fields, packages)
    tb.align['NAME'] = 'l'
    tb.align['VERSION'] = 'l'
    tb.align['SELECTED'] = 'l'
    tb.align['FRAMEWORK'] = 'l'
    tb.align['DESCRIPTION'] = 'l'

    return tb


def auth_provider_table(providers):
    """Returns a PrettyTable representation of the auth providers for cluster

    :param providers: auth providers available
    :type providers: dict
    :rtype: PrettyTable

    """

    fields = OrderedDict([
        ('PROVIDER ID', lambda p: p),
        ('AUTHENTICATION TYPE', lambda p: auth.auth_type_description(
                                            providers[p])),
    ])

    tb = table(fields, providers, sortby="PROVIDER ID")
    tb.align['PROVIDER ID'] = 'l'
    tb.align['AUTHENTICATION TYPE'] = 'l'

    return tb


def clusters_table(clusters):
    """Returns a PrettyTable representation of the configured clusters

    :param clusters: configured clusters
    :type clusters: [Cluster]
    :rtype: PrettyTable
    """

    def print_name(c):
        msg = c['name']
        if c['attached']:
            msg += "*"
        return msg

    fields = OrderedDict([
        ('NAME', lambda c: print_name(c)),
        ('CLUSTER ID', lambda c: c['cluster_id']),
        ('VERSION', lambda c: c['version']),
        ('URL', lambda c: c['url'] or "N/A")
    ])

    tb = table(fields, clusters, sortby="CLUSTER ID")

    return tb


def node_table(nodes, field_names=()):
    """Returns a PrettyTable representation of the provided DC/OS nodes

    :param nodes: nodes to render.
    :type nodes: [dict]
    :param field_names: Extra fields to add to the table
    :type nodes: [str]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ('HOSTNAME', lambda s: s.get('host', s.get('hostname'))),
        ('IP', lambda s: s.get('ip') or mesos.parse_pid(s['pid'])[1]),
        ('ID', lambda s: s['id']),
        ('TYPE', lambda s: s['type']),
    ])

    for field_name in field_names:
        if field_name.upper() in fields:
            continue
        if ':' in field_name:
            heading, field_name = field_name.split(':', 1)
        else:
            heading = field_name
        fields[heading.upper()] = _dotted_itemgetter(field_name)

    sortby = list(fields.keys())[0]
    tb = table(fields, nodes, sortby=sortby)
    tb.align['TYPE'] = 'l'
    return tb


def _dotted_itemgetter(field_name):
    """Returns a func that gets the value in a nested dict where the
    `field_name` is a dotted path to the key.

    Example:

      >>> from dcoscli.tables import _dotted_itemgetter
      >>> d1 = {'a': {'b': {'c': 21}}}
      >>> d2 = {'a': {'b': {'c': 22}}}
      >>> func = _dotted_itemgetter('a.b.c')
      >>> func(d1)
      21
      >>> func(d2)
      22

    :param field_name: dotted path to key in nested dict
    :type field_name: str
    :rtype: callable
    """

    if '.' not in field_name:
        return operator.itemgetter(field_name)
    head, tail = field_name.split('.', 1)
    return lambda d: _dotted_itemgetter(tail)(d[head])


def _format_unix_timestamp(ts):
    """ Formats a unix timestamp in a `dcos task ls --long` format.

    :param ts: unix timestamp
    :type ts: int
    :rtype: str
    """
    return datetime.datetime.fromtimestamp(ts).strftime('%b %d %H:%M')


def ls_long_table(files):
    """Returns a PrettyTable representation of `files`

    :param files: Files to render.  Of the form returned from the
        mesos /files/browse.json endpoint.
    :param files: [dict]
    :rtype: PrettyTable
    """

    fields = OrderedDict([
        ('MODE', lambda f: f['mode']),
        ('NLINK', lambda f: f['nlink']),
        ('UID', lambda f: f['uid']),
        ('GID', lambda f: f['gid']),
        ('SIZE', lambda f: f['size']),
        ('DATE', lambda f: _format_unix_timestamp(int(f['mtime']))),
        ('PATH', lambda f: posixpath.basename(f['path']))])

    tb = table(fields, files, sortby="PATH", header=False)
    tb.align = 'r'
    return tb


def metrics_summary_table(data):
    """Prints a table of CPU, Memory and Disk for the given data.

    :param data: A dictionary of formatted summary values.
    :type data: dict
    :rtype: PrettyTable
    """
    fields = OrderedDict([
        ('CPU', lambda d: d['cpu']),
        ('MEM', lambda d: d['mem']),
        ('DISK', lambda d: d['disk'])
    ])

    # table has a single row
    metrics_table = table(fields, [data])
    metrics_table.align['CPU'] = 'l'
    metrics_table.align['MEM'] = 'l'
    metrics_table.align['DISK'] = 'l'

    return metrics_table


def metrics_details_table(datapoints, show_tags=True):
    """Prints a table of all passed metrics

    :param datapoints: A raw list of datapoints
    :type datapoints: [dict]
    :param show_tags: Show column for tags, unless False
    :type show_tags: bool
    :rtype: PrettyTable
    """

    field_defs = [
        ('NAME', lambda d: d['name']),
        ('VALUE', lambda d: d['value']),
    ]
    if show_tags:
        field_defs.append(('TAGS', lambda d: d['tags']))

    fields = OrderedDict(field_defs)

    metrics_table = table(fields, datapoints)
    for (k, v) in field_defs:
        metrics_table.align[k] = 'l'

    return metrics_table


def truncate_table(fields, objs, limits, **kwargs):
    """Returns a PrettyTable.  `fields` represents the header schema of
    the table.  `objs` represents the objects to be rendered into
    rows.

    :param fields: An OrderedDict, where each element represents a
                   column.  The key is the column header, and the
                   value is the function that transforms an element of
                   `objs` into a value for that column.
    :type fields: OrderdDict(str, function)
    :param objs: objects to render into rows
    :type objs: [object]
    :param limits: limits for truncating for each row
    :type limits: [object]
    :param **kwargs: kwargs to pass to `prettytable.PrettyTable`
    :type **kwargs: dict
    :rtype: PrettyTable
    """

    tb = prettytable.PrettyTable(
        [k.upper() for k in fields.keys()],
        border=False,
        hrules=prettytable.NONE,
        vrules=prettytable.NONE,
        left_padding_width=0,
        right_padding_width=1,
        **kwargs
    )

    # Set these explicitly due to a bug in prettytable where
    # '0' values are not honored.
    tb._left_padding_width = 0
    tb._right_padding_width = 2

    def format_table(obj, key, function):
        """Formats the given object for the given function

        :param object: object to format
        :type object: object
        :param key: value which should be checked
        :type key: string
        :param function: function to format the cell
        :type function: function
        :rtype: PrettyTable
        """
        try:
            result = str(function(obj))
        except KeyError:
            result = 'N/A'
        if (limits is not None and limits.get(key) is not None):
            result = textwrap.\
                shorten(result, width=limits.get(key), placeholder='...')
        return result

    for obj in objs:
        row = [format_table(obj, key, fields.get(key))
               for key in fields.keys()]
        tb.add_row(row)

    return tb


def table(fields, objs, **kwargs):
    """Returns a PrettyTable.  `fields` represents the header schema of
    the table.  `objs` represents the objects to be rendered into
    rows.

    :param fields: An OrderedDict, where each element represents a
                   column.  The key is the column header, and the
                   value is the function that transforms an element of
                   `objs` into a value for that column.
    :type fields: OrderdDict(str, function)
    :param objs: objects to render into rows
    :type objs: [object]
    :param **kwargs: kwargs to pass to `prettytable.PrettyTable`
    :type **kwargs: dict
    :rtype: PrettyTable
    """

    return truncate_table(fields, objs, None, **kwargs)



from sseclient import SSEClient

from . import http


def get(url, **kwargs):
    """ Make a get request to streaming endpoint which
    implements SSE (Server sent events). The parameter session=http
    will ensure we are using `dcos.http` module with all required auth
    headers.

    :param url: server sent events streaming URL
    :type url: str
    :param kwargs: arbitrary params for requests
    :type kwargs: dict
    :return: instance of sseclient.SSEClient
    :rtype: sseclient.SSEClient
    """
    return SSEClient(url, session=http, **kwargs)



import os

from ..integrations.helpers.common import file_json_ast

PODS_FILE_PATH_BASE = 'tests/data/marathon/pods'

GOOD_POD_ID = 'good-pod'
GOOD_POD_FILE_PATH = os.path.join(PODS_FILE_PATH_BASE, 'good.json')
UNGOOD_POD_FILE_PATH = \
    os.path.join(PODS_FILE_PATH_BASE, 'ungood.json')
UPDATED_GOOD_POD_FILE_PATH = \
    os.path.join(PODS_FILE_PATH_BASE, 'updated_good.json')

DOUBLE_POD_ID = 'double-pod'
DOUBLE_POD_FILE_PATH = os.path.join(PODS_FILE_PATH_BASE, 'double.json')

TRIPLE_POD_ID = 'winston'
TRIPLE_POD_FILE_PATH = os.path.join(PODS_FILE_PATH_BASE, 'doubleplusgood.json')

POD_KILL_ID = 'pod-kill'
POD_KILL_FILE_PATH = os.path.join(PODS_FILE_PATH_BASE,
                                  'pod_kill.json')


def app_fixture():
    """ Marathon app fixture.

    :rtype: dict
    """

    return {
        "acceptedResourceRoles": None,
        "args": None,
        "backoffFactor": 1.15,
        "backoffSeconds": 1,
        "cmd": "sleep 1000",
        "constraints": [],
        "container": None,
        "cpus": 0.1,
        "dependencies": [],
        "deployments": [],
        "waiting": "False",
        "disk": 0.0,
        "env": {},
        "executor": "",
        "healthChecks": [],
        "id": "/test-app",
        "instances": 1,
        "labels": {
            "PACKAGE_ID": "test-app",
            "PACKAGE_VERSION": "1.2.3"
        },
        "maxLaunchDelaySeconds": 3600,
        "mem": 16.0,
        "ports": [
            10000
        ],
        "requirePorts": False,
        "storeUrls": [],
        "tasksHealthy": 0,
        "tasksRunning": 1,
        "tasksStaged": 0,
        "tasksUnhealthy": 0,
        "upgradeStrategy": {
            "maximumOverCapacity": 1.0,
            "minimumHealthCapacity": 1.0
        },
        "uris": [],
        "user": None,
        "version": "2015-05-28T21:21:05.064Z"
    }


def deployment_fixture_app_pre_pods():
    """ Marathon app deployment fixture, for Marathon versions without pods.

    :rtype: dict
    """

    return {
        "affectedApps": [
            "/cassandra/dcos"
        ],
        "currentActions": [
            {
                "action": "ScaleApplication",
                "app": "/cassandra/dcos"
            }
        ],
        "currentStep": 2,
        "id": "bebb8ffd-118e-4067-8fcb-d19e44126911",
        "steps": [
            [
                {
                    "action": "StartApplication",
                    "app": "/cassandra/dcos"
                }
            ],
            [
                {
                    "action": "ScaleApplication",
                    "app": "/cassandra/dcos"
                }
            ]
        ],
        "totalSteps": 2,
        "version": "2015-05-29T01:13:47.694Z"
    }


def deployment_fixture_app_post_pods():
    """ Marathon app deployment fixture, for Marathon versions with pods.

    :rtype: dict
    """

    return {
        "affectedApps": [
            "/cassandra/dcos"
        ],
        "affectedPods": [],
        "currentActions": [
            {
                "action": "ScaleApplication",
                "app": "/cassandra/dcos"
            }
        ],
        "currentStep": 2,
        "id": "bebb8ffd-118e-4067-8fcb-d19e44126911",
        "steps": [
            [
                {
                    "action": "StartApplication",
                    "app": "/cassandra/dcos"
                }
            ],
            [
                {
                    "action": "ScaleApplication",
                    "app": "/cassandra/dcos"
                }
            ]
        ],
        "totalSteps": 2,
        "version": "2015-05-29T01:13:47.694Z"
    }


def deployment_fixture_pod():
    """ Marathon pod deployment fixture.

    :rtype: dict
    """

    return {
        "affectedApps": [],
        "affectedPods": [
            "/cassandra/dcos"
        ],
        "currentActions": [
            {
                "action": "ScalePod",
                "pod": "/cassandra/dcos"
            }
        ],
        "currentStep": 2,
        "id": "bebb8ffd-118e-4067-8fcb-d19e44126911",
        "steps": [
            [
                {
                    "action": "StartPod",
                    "pod": "/cassandra/dcos"
                }
            ],
            [
                {
                    "action": "ScalePod",
                    "pod": "/cassandra/dcos"
                }
            ]
        ],
        "totalSteps": 2,
        "version": "2015-05-29T01:13:47.694Z"
    }


def app_task_fixture():
    """ Marathon task fixture.

    :rtype: dict
    """

    return {
        "appId": "/zero-instance-app",
        "host": "dcos-01",
        "id": "zero-instance-app.027b3a83-063d-11e5-84a3-56847afe9799",
        "ports": [
            8165
        ],
        "servicePorts": [
            10001
        ],
        "stagedAt": "2015-05-29T19:58:00.907Z",
        "startedAt": "2015-05-29T19:58:01.114Z",
        "version": "2015-05-29T18:50:58.941Z"
    }


def group_fixture():
    """ Marathon group fixture.

    :rtype: dict
    """

    return {
        "apps": [],
        "dependencies": [],
        "groups": [
            {
                "apps": [
                    {
                        "acceptedResourceRoles": None,
                        "args": None,
                        "backoffFactor": 1.15,
                        "backoffSeconds": 1,
                        "cmd": "sleep 1",
                        "constraints": [],
                        "container": None,
                        "cpus": 1.0,
                        "dependencies": [],
                        "disk": 0.0,
                        "env": {},
                        "executor": "",
                        "healthChecks": [],
                        "id": "/test-group/sleep/goodnight",
                        "instances": 0,
                        "labels": {},
                        "maxLaunchDelaySeconds": 3600,
                        "mem": 128.0,
                        "ports": [
                            10000
                        ],
                        "requirePorts": False,
                        "storeUrls": [],
                        "upgradeStrategy": {
                            "maximumOverCapacity": 1.0,
                            "minimumHealthCapacity": 1.0
                        },
                        "uris": [],
                        "user": None,
                        "version": "2015-05-29T23:12:46.187Z"
                    }
                ],
                "dependencies": [],
                "groups": [],
                "id": "/test-group/sleep",
                "version": "2015-05-29T23:12:46.187Z"
            }
        ],
        "id": "/test-group",
        "version": "2015-05-29T23:12:46.187Z"
    }


def pod_list_fixture():
    """"Marathon pod list fixture.

    :rtype: [{}]
    """

    good_pod_spec = file_json_ast(GOOD_POD_FILE_PATH)
    double_pod_spec = file_json_ast(DOUBLE_POD_FILE_PATH)
    triple_pod_spec = file_json_ast(TRIPLE_POD_FILE_PATH)

    good_pod_spec["version"] = "2016-10-05T16:18:03.399Z"
    double_pod_spec["version"] = "2016-10-04T27:18:28.183Z"
    triple_pod_spec["version"] = "2016-09-01T31:41:59.265Z"

    good_pod_status = {
        "id": good_pod_spec["id"],
        "instances": [{}, {}, {}],
        "spec": good_pod_spec,
        "status": "stable",
        "statusSince": "2016-10-05T15:36:54.123Z",
        "waiting": "False"
    }

    double_pod_status = {
        "id": double_pod_spec["id"],
        "instances": [{}, {}],
        "spec": double_pod_spec,
        "status": "terminal",
        "statusSince": "2016-10-04T03:13:37.101Z",
        "waiting": "False"
    }

    triple_pod_status = {
        "id": triple_pod_spec["id"],
        "instances": [{}],
        "spec": triple_pod_spec,
        "status": "degraded",
        "statusSince": "2016-09-30T22:28:09.201Z",
        "waiting": "False"
    }

    return [good_pod_status, double_pod_status, triple_pod_status]


def pod_list_without_instances_fixture():
    """Marathon pod without an "instances" field.

    :rtype: {}
    """

    return {
        "id": "/pod-without-instances",
        "spec": {
            "id": "/pod-without-instances",
            "containers": [
                {
                    "name": "no-instances",
                    "exec": {
                        "command": {
                            "shell": "sleep 1000"
                        }
                    },
                    "resources": {
                        "cpus": 0.1,
                        "mem": 16.0
                    }
                }
            ],
            "networks": [
                {
                    "mode": "host"
                }
            ],
            "version": "2016-10-05T16:18:03.399Z"
        },
        "status": "stable",
        "statusSince": "2016-10-05T15:36:54.123Z"
    }


def pod_list_without_spec_version_fixture():
    """Marathon pod without a "spec"."version" field.

    :rtype: {}
    """

    return {
        "id": "/pod-without-spec-version",
        "instances": [{}],
        "spec": {
            "id": "/pod-without-spec-version",
            "containers": [
                {
                    "name": "no-spec-version",
                    "exec": {
                        "command": {
                            "shell": "sleep 1000"
                        }
                    },
                    "resources": {
                        "cpus": 0.1,
                        "mem": 16.0
                    }
                }
            ],
            "networks": [
                {
                    "mode": "host"
                }
            ]
        },
        "status": "stable",
        "statusSince": "2016-10-05T15:36:54.123Z"
    }



import six


def assert_same_elements(list1, list2):
    """Tests whether the second argument is a rearrangement of the first.

    This method will work even if the list elements are mutable and/or cannot
    be sorted. A common use case is comparing lists of JSON objects, since
    the objects are usually implemented with dicts.

    :param list1: the first list
    :type: list
    :param list2: the second list
    :type: list
    :rtype: None
    """
    list2 = list2.copy()
    for element in list1:
        list2.remove(element)
    assert list2 == []


def file_bytes(path):
    """ Read all bytes from a file

    :param path: path to file
    :type path: str
    :rtype: bytes
    :returns: bytes from the file
    """

    with open(path) as f:
        return six.b(f.read())



import mock
import pytest

import dcoscli.node.main as main
from dcos.errors import DCOSException


@mock.patch('dcos.packagemanager.PackageManager')
def test_check_version_fail(mock_cosmos):
    """
    Test _check_3dt_version(), should throw DCOSException exception.
    """

    mock_cosmos().enabled.return_value = True
    mock_cosmos().has_capability.return_value = False

    with pytest.raises(DCOSException) as excinfo:
        main._check_3dt_version()
    assert str(excinfo.value) == (
        'DC/OS backend does not support diagnostics capabilities in this '
        'version. Must be DC/OS >= 1.8')


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
def test_check_version_success(mock_get, mock_cosmos):
    """
    Test _check_3dt_version(), should not fail.
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    main._check_3dt_version()


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.node.main._do_diagnostics_request')
def test_node_diagnostics_create(mock_do_diagnostics_request, mock_get,
                                 mock_cosmos):
    """
    Test _bundle_create(), should not fail.
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    mock_do_diagnostics_request.return_value = {
        'status': 'OK',
        'extra': {
            'bundle_name': 'bundle.zip'
        }
    }
    main._bundle_create(['10.10.0.1'])
    mock_do_diagnostics_request.assert_called_once_with(
        '/system/health/v1/report/diagnostics/create',
        'POST',
        json={'nodes': ['10.10.0.1']})


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.node.main._do_diagnostics_request')
def test_node_diagnostics_delete(mock_do_diagnostics_request, mock_get,
                                 mock_cosmos):
    """
    Test _bundle_delete(), should not fail
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    mock_do_diagnostics_request.return_value = {
        'status': 'OK'
    }
    main._bundle_delete('bundle.zip')
    mock_do_diagnostics_request.assert_called_once_with(
        '/system/health/v1/report/diagnostics/delete/bundle.zip',
        'POST'
    )


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.node.main._do_diagnostics_request')
def test_node_diagnostics_list(mock_do_diagnostics_request, mock_get,
                               mock_cosmos):
    """
    Test _bundle_manage(), should not fail
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    mock_do_diagnostics_request.return_value = {
        '127.0.0.1': [
            {
                'file_name': 'bundle.zip',
                'file_size': 123
            }
        ]
    }

    # _bundle_manage(list_bundles, status, cancel, json)
    main._bundle_manage(True, False, False, False)
    mock_do_diagnostics_request.assert_called_once_with(
        '/system/health/v1/report/diagnostics/list/all',
        'GET'
    )


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.node.main._do_diagnostics_request')
def test_node_diagnostics_status(mock_do_diagnostics_request, mock_get,
                                 mock_cosmos):
    """
    Test _bundle_manage(), should not fail
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    mock_do_diagnostics_request.return_value = {
        'host1': {
            'prop1': 'value1'
        }
    }

    # _bundle_manage(list_bundles, status, cancel, json)
    main._bundle_manage(False, True, False, False)
    mock_do_diagnostics_request.assert_called_once_with(
        '/system/health/v1/report/diagnostics/status/all',
        'GET'
    )


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.node.main._do_diagnostics_request')
def test_node_diagnostics_cancel(mock_do_diagnostics_request, mock_get,
                                 mock_cosmos):
    """
    Test _bundle_manage(), should not fail
    """

    mock_cosmos().enabled.return_value = True
    m = mock.MagicMock()
    m.json.return_value = {
        'capabilities': [{'name': 'SUPPORT_CLUSTER_REPORT'}]
    }
    mock_get.return_value = m
    mock_do_diagnostics_request.return_value = {
        'status': 'success'
    }

    # _bundle_manage(list_bundles, status, cancel, json)
    main._bundle_manage(False, False, True, False)
    mock_do_diagnostics_request.assert_called_once_with(
        '/system/health/v1/report/diagnostics/cancel',
        'POST'
    )


@mock.patch('dcos.packagemanager.PackageManager')
@mock.patch('dcoscli.node.main._do_request')
@mock.patch('dcoscli.node.main._get_bundle_list')
def test_node_diagnostics_download(mock_get_diagnostics_list, mock_do_request,
                                   mock_cosmos):
    mock_cosmos().enabled.return_value = True
    mock_get_diagnostics_list.return_value = [('bundle.zip', 123)]
    main._bundle_download('bundle.zip', None)
    mock_do_request.assert_called_with(
        '/system/health/v1/report/diagnostics/serve/bundle.zip', 'GET',
        stream=True)


@mock.patch('dcos.config.get_config_val')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.log.dcos_log_enabled')
def test_dcos_log_leader_mesos(mocked_dcos_log_enabked, mocked_http_get,
                               mocked_get_config_val):
    mocked_dcos_log_enabked.return_value = True

    m = mock.MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    mocked_get_config_val.return_value = 'http://127.0.0.1'

    main._dcos_log(False, 10, True, '', None, [])
    mocked_http_get.assert_called_with(
        'http://127.0.0.1/system/v1/leader/mesos/logs/v1/range/?skip_prev=10',
        headers={'Accept': 'text/plain'})


@mock.patch('dcos.config.get_config_val')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.log.dcos_log_enabled')
def test_dcos_log_leader_marathon(mocked_dcos_log_enabked, mocked_http_get,
                                  mocked_get_config_val):
    mocked_dcos_log_enabked.return_value = True

    m = mock.MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    mocked_get_config_val.return_value = 'http://127.0.0.1'

    main._dcos_log(False, 10, True, '', 'dcos-marathon', [])
    mocked_http_get.assert_called_with(
        'http://127.0.0.1/system/v1/leader/marathon/logs/v1/range/'
        '?skip_prev=10&filter=_SYSTEMD_UNIT:dcos-marathon.service',
        headers={'Accept': 'text/plain'})


@mock.patch('dcoscli.log.follow_logs')
@mock.patch('dcos.config.get_config_val')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.log.dcos_log_enabled')
def test_dcos_log_stream(mocked_dcos_log_enabked, mocked_http_get,
                         mocked_get_config_val, mocked_follow_logs):
    mocked_dcos_log_enabked.return_value = True

    m = mock.MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    mocked_get_config_val.return_value = 'http://127.0.0.1'

    main._dcos_log(True, 20, False, 'mesos-id', None, [])
    mocked_follow_logs.assert_called_with(
        'http://127.0.0.1/system/v1/agent/mesos-id/logs'
        '/v1/stream/?skip_prev=20')


@mock.patch('dcoscli.log.follow_logs')
@mock.patch('dcos.config.get_config_val')
@mock.patch('dcos.http.get')
@mock.patch('dcoscli.log.dcos_log_enabled')
def test_dcos_log_filters(mocked_dcos_log_enabked, mocked_http_get,
                          mocked_get_config_val, mocked_follow_logs):
    mocked_dcos_log_enabked.return_value = True

    m = mock.MagicMock()
    m.status_code = 200
    mocked_http_get.return_value = m

    mocked_get_config_val.return_value = 'http://127.0.0.1'

    main._dcos_log(True, 20, False, 'mesos-id', 'dcos-mesos-master',
                   ['key1:value1', 'key2:value2'])

    mocked_follow_logs.assert_called_with(
        'http://127.0.0.1/system/v1/agent/mesos-id/logs/v1/stream/'
        '?skip_prev=20&filter=key1:value1&filter=key2:value2&'
        'filter=_SYSTEMD_UNIT:dcos-mesos-master.service')


@mock.patch('dcos.config.get_config_val')
@mock.patch('dcoscli.node.main._get_slave_ip')
@mock.patch('dcos.http.get')
def test_list_components(mocked_get, mocked_get_slave_ip,
                         mocked_get_config_val):
    m = mock.MagicMock()
    m.json.return_value = {
        'units': [
            {
                'id': 'dcos-component.service',
            }
        ]
    }
    mocked_get.return_value = m
    mocked_get_slave_ip.return_value = '127.0.0.1'
    mocked_get_config_val.return_value = 'http://10.10.10.10'
    main._list_components(None, 'slave-id', False)
    mocked_get.assert_called_with(
        'http://10.10.10.10/system/health/v1/nodes/127.0.0.1/units')


@mock.patch('dcos.config.get_config_val')
@mock.patch('dcos.mesos.MesosDNSClient')
@mock.patch('dcos.http.get')
def test_list_components_leader(mocked_get, mocked_dns,
                                mocked_get_config_val):
    m = mock.MagicMock()
    m.json.return_value = {
        'units': [
            {
                'id': 'dcos-component.service',
            }
        ]
    }
    mocked_dns().hosts.return_value = [{'ip': '10.10.0.1'}]
    mocked_get_config_val.return_value = 'http://10.10.10.10'

    mocked_get.return_value = m
    main._list_components(True, False, False)
    mocked_get.assert_called_with(
        'http://10.10.10.10/system/health/v1/nodes/10.10.0.1/units')
